@@ -15,8 +15,8 @@
 #include <pwd.h>
 #include "mail.h"
 #include "zmsignal.h"
-#ifdef FCNTL_H
-# include FCNTL_H
+#ifdef HAVE_FCNTL_H
+# include <fcntl.h>
 #endif
 #include <sys/file.h>
 #ifdef	HAVE_SYSLOG
@@ -39,6 +39,7 @@ extern char *optarg;
 extern char *strchr();
 extern char *strrchr();
 #endif
+extern char *strtok();
 extern char *getenv();
 extern char *getlogin();
 extern int errno;
