@@ -42,6 +42,6 @@ extern int getmyhostname __((char *namebuf, int len));
 /* setreuid ?? SysV beastie.. */
 
 /* setvbuf.c */
-#ifndef USE_SETVBUF
+#ifndef HAVE_SETVBUF
 exten int setvbuf __((FILE *fp, char *buf, int type, int size));
 #endif
