@@ -165,7 +165,7 @@ extern int ermdir __((char *path));
 extern int erename __((char *from, char *to));
 
 /* lockaddr.c: */
-extern int lockaddr __((int fd, char *map, long offset, int was, int new));
+extern int lockaddr __((int fd, char *map, long offset, int was, int new, const char *file, const char *host));
 
 /* markoff.c: */
 extern int markoff __((char *filecontents, int bytesleft, long offsets[], char *filename));
