@@ -10,8 +10,8 @@
 #include <errno.h>
 #include <sys/param.h>
 #include <sys/stat.h>
-#ifdef FCNTL_H
-# include FCNTL_H
+#ifdef HAVE_FCNTL_H
+# include <fcntl.h>
 #endif
 #include <sys/file.h>
 #include <sys/socket.h>
