@@ -468,7 +468,7 @@ __p_rr(cp, msg, file)
 		switch (class) {
 		case C_IN:
 		case C_HS:
-			bcopy(cp, (char *)&inaddr, INADDRSZ);
+			memcpy((char *)&inaddr, cp, INADDRSZ);
 			if (dlen == 4) {
 				fprintf(file, "\t%s", inet_ntoa(inaddr));
 				cp += dlen;
@@ -635,7 +635,7 @@ __p_rr(cp, msg, file)
 	case T_WKS:
 		if (dlen < INT32SZ + 1)
 			break;
-		bcopy(cp, (char *)&inaddr, INADDRSZ);
+		memcpy((char *)&inaddr, cp, INADDRSZ);
 		cp += INT32SZ;
 		fprintf(file, "\t%s %s ( ",
 			inet_ntoa(inaddr),
