@@ -53,7 +53,10 @@ int std_puts(s) char *s; { return puts(s); }
 int std_fputs(s, fp) char *s; FILE *fp; { return fputs(s, fp); }
 
 #ifdef HAVE_STDARG_H
-int (*std_printf) __((char *fmt, ...)) = (int (*) __((char *fmt, ...)))printf;
+#if (defined(__GNUC__) && defined(sun) && !defined(__svr4__))
+extern int printf(); /* Sun 4.1.x with old GCC */
+#endif
+int (*std_printf) __((char *fmt, ...)) = (int(*)())printf;
 #else
 int (*std_printf) __((/*char *fmt, ...*/)) = (int (*) __((/* char *fmt, ... */)))printf;
 #endif
