@@ -16,8 +16,8 @@
 #include "mailer.h"
 #include <ctype.h>
 #include <sys/file.h>
-#ifdef FCNTL_H
-# include FCNTL_H
+#ifdef HAVE_FCNTL_H
+# include <fcntl.h>
 #endif
 #include "libdb/search.h"
 #include "splay.h"
