@@ -322,7 +322,7 @@ modp_gdbm(sip)
 	spl = sp_lookup(symid, spt_modcheck);
 	if (spl != NULL) {
 		rval = stbuf.st_mtime != (time_t)spl->data
-			|| stbuf.st_nlink != (int)spl->mark
+			|| (long)stbuf.st_nlink != (long)spl->mark
 			|| stbuf.st_nlink == 0;
 	} else
 		rval = 0;
