@@ -30,12 +30,18 @@ extern char	*prio_list[];
 extern char	*progname;
 extern char	*rfc822tz __(());
 extern char	*sb_retrieve __(());
+#ifndef HAVE_STRERROR /* System has it, and propably has prototype too..
+			 IRIX 6.2 */
 extern char	*strerror __((const int errno));
+#endif
 extern int	D_hdr_rewrite;
 extern int	D_router;
 extern int	cistrcmp();
 extern int	cistrncmp();
-extern char	*strchr(), *strrchr();
+#ifndef strchr
+extern char	*strchr();
+extern char	*strrchr();
+#endif
 extern int	strmatch();
 extern int	apply();
 extern int	lapply();
@@ -134,9 +140,8 @@ extern void	dumpHeaders __((struct header *h));
 extern void	dumpHeader __((struct header *h));
 extern int	isSenderAddr;
 extern int	isRecpntAddr;
-extern void	squirrel __((struct envelope *e, char *text));
+extern void	squirrel __((struct envelope *e, char *keyword, char *text));
 extern struct header	*erraddress __((struct envelope *e));
-extern void	reject __((struct envelope *e, char *msgfile));
 extern void	defer __((struct envelope *e, char *why));
 extern struct header	*mkSender __((struct envelope *e, u_char *name, int flag));
 extern struct header	*mkTrace __((struct envelope *e));
