@@ -594,6 +594,35 @@ erraddress(e)
 	return h;
 }
 
+
+/* Pick recipient address from the input line.
+   EXTREMELY Simple minded parsing.. */
+static void pick_env_addr(buf,mfp)
+char *buf;
+FILE *mfp;
+{
+	char *s = buf;
+
+	while (*s != 0 && *s != ' ' && *s != '\t' && *s != ':') ++s;
+	if (*s != ':') return; /* BAD INPUT! */
+	buf = ++s; /* We have skipped the initial header.. */
+
+	s = strchr(buf,'\n');
+	if (s) *s = 0;
+	s = strchr(buf,'<');
+	if (s != NULL) {
+	  /*  Cc:  The Postoffice managers <postoffice> */
+	  buf = ++s;
+	  s = strrchr(buf,'>');
+	  if (s) *s = 0;
+	  else return; /* No trailing '>' ? BAD BAD! */
+	  fprintf(mfp,"to <%s>\n",buf);
+	} else {
+	  /*  Cc: some-address  */
+	  fprintf(mfp,"to <%s>\n",buf);
+	}
+}
+
 /* Send the message back to originator with user-friendly chastizing errors */
 
 static void
@@ -638,18 +667,31 @@ reject(e, msgfile)
 		squirrel(e, "mailcreatefail", "Couldn't open reject message file");
 		return;
 	}
+	setvbuf(fp, vbuf, _IOFBF, sizeof vbuf);
+	while (fgets(buf,sizeof(buf),fp) != NULL) {
+	  int rc;
+	  if (strncmp("ADR",buf,3)==0) {
+	    pick_env_addr(buf+4,mfp);
+	  } else if (strncmp("HDR",buf,3)==0 ||
+		     strncmp("SUB",buf,3)==0) {
+	    continue;
+	  } else
+	    break;
+	}
+	fputs("env-end\n",mfp);
+	fseek(fp,0,0);
+
 	/* who is it from? */
 	cp = h->h_pname;
 	h->h_pname = "To";
 	hdr_print(h, mfp);
-	/* XX: if forcing postmaster cc:, do it here */
 	h->h_pname = cp;
 	lastCh = nextToLastCh = '\0';
-	setvbuf(fp, vbuf, _IOFBF, sizeof vbuf);
 	while (fgets(buf,sizeof(buf),fp) != NULL) {
 	  int rc;
 	  n = strlen(buf);
 	  if (strncmp("HDR",buf,3)==0 ||
+	      strncmp("ADR",buf,3)==0 ||
 	      strncmp("SUB",buf,3)==0) {
 	    rc = fputs(buf+4,mfp);
 	    n -= 4;
