@@ -332,7 +332,7 @@ ok:
  * tEndOfHeader then emit the ePrematureEndOfFile error code and terminate
  * execution.  Otherwise, emit the eSyntaxError error code and set the
  * nextInputToken to tSyntaxError to prevent further input, and exit when the
- * expected (??) input is tSemicolon or tNewLine.
+ * expected ( ?? ) input is tSemicolon or tNewLine.
  *
  * If the expected token is not tSemicolon nor tNewLine and a syntax error
  * has already been detected on the current line (flagged by nextInputToken ==
@@ -705,7 +705,7 @@ parse822(entry, tlistp, ltmp, tfp)
 
 	if (tracefp) {
 	  sslPointer = ((int)entry) + 1;
-	  operation = -1;
+	  operation = (TableOperation)-1; /* Invalid value! */
 	  SslTrace();
 	}
 
