@@ -135,11 +135,18 @@ msgerror(vp, offset, message)
 {
 	FILE *fp;
 	char *notary = "";
+	char path[128];
 
 	if (vp->notary) notary = vp->notary;
 
+	if (vp->cfp->dirind > 0) {
+	  sprintf(path,"%s/%s", cfpdirname(vp->cfp->dirind), vp->cfp->mid);
+	} else {
+	  strcpy(path,vp->cfp->mid);
+	}
+
 	/* exclusive access required, but we're the only scheduler... */
-	if ((fp = fopen(vp->cfp->mid, "a")) == NULL) {
+	if ((fp = fopen(path, "a")) == NULL) {
 	  fprintf(stderr,
 		  "Cannot open control file %s to deposit", vp->cfp->mid);
 	  fprintf(stderr,
@@ -193,7 +200,7 @@ reporterrs(cfpi)
 {
 	int i, n, wroteheader, byteidx, headeridx = -1, drptidx, fd;
 	long *lp;
-	char path[MAXPATHLEN], buf[BUFSIZ], *midbuf, *cp, *s, *eaddr;
+	char *midbuf, *cp, *s, *eaddr;
 	char *deliveryform;
 	FILE *errfp, *fp;
 	char *boundarystr;
@@ -206,6 +213,7 @@ reporterrs(cfpi)
 	struct ctlfile *cfp;
 	char *rcpntpointer;
 	int no_error_report = 0;
+	char path[MAXPATHLEN], buf[BUFSIZ], mpath[MAXPATHLEN];
 
 	if (cfpi->haderror == 0)
 		return;
@@ -224,11 +232,16 @@ reporterrs(cfpi)
 	midbuf = cfpi->mid;
 
 	/* exclusive access required, but we're the only scheduler... */
-	if ((fd = open(cfpi->mid, O_RDWR, 0)) < 0 ||
+	if (cfpi->dirind > 0)
+	  sprintf(mpath,"%s/%s", cfpdirname(cfpi->dirind), cfpi->mid);
+	else
+	  strcpy(mpath,cfpi->mid);
+
+	if ((fd = open(mpath, O_RDWR, 0)) < 0 ||
 	    (cfp = slurp(fd, cfpi->id)) == NULL) {
 	  fprintf(stderr,
 		  "%s: unexpected absence of control file %s for error processing!\n",
-		  progname, cfpi->mid);
+		  progname, mpath);
 	  if (fd >= 0)
 	    close(fd);
 	  return;
@@ -285,6 +298,8 @@ reporterrs(cfpi)
 
 	      if (no_error_report == 0)
 		fprintf(errfp, "To: %s\n", eaddr);
+	      else if (no_error_report < 0)
+		fprintf(errfp, "To: dummy:; (error trapped source)\n");
 	      while (fgets(buf,sizeof(buf)-1,fp) != NULL) {
 		if (strncmp(buf,"HDR",3)==0) {
 		  fputs(buf+4,errfp);
@@ -382,7 +397,7 @@ reporterrs(cfpi)
 	    putc('\n', errfp);
 	  } else
 	    fprintf(errfp, "%s\n", cp);
-	  lockaddr(cfp->fd, NULL, *lp + 1, _CFTAG_NORMAL, _CFTAG_OK);
+	  lockaddr(cfp->fd, NULL, *lp + 1, _CFTAG_NORMAL, _CFTAG_OK, mpath, "diagmsg");
 #ifdef	LOG_INFO
 	  syslog(LOG_INFO, "%s: <%s>: %s",
 		 cfp->mid, cfp->contents + byteidx + 2, cp);
@@ -415,7 +430,14 @@ reporterrs(cfpi)
 	fprintf(errfp, "--%s\n", boundarystr);
 	fprintf(errfp, "Content-Type: message/rfc822\n\n");
 
-	sprintf(path, "../%s/%s", QUEUEDIR, cfp->mid);
+	/* path to the message body */
+	if (cfpi->dirind > 0)
+	  sprintf(path, "../%s/%s/%s",
+		  QUEUEDIR, cfpdirname(cfpi->dirind), cfpi->mid);
+	else
+	  sprintf(path, "../%s/%s",
+		  QUEUEDIR, cfpi->mid);
+
 	if ((fp = fopen(path, "r")) != NULL) {
 
 	  if (cfp->msgbodyoffset > 0 && headeridx > 0 ) {
@@ -447,6 +469,8 @@ reporterrs(cfpi)
 	  while ((n = fread(buf, sizeof buf[0], sizeof buf, fp)) > 0)
 	    fwrite(buf, sizeof buf[0], n, errfp);
 	  fclose(fp);
+	} else {
+	  fprintf(stderr,"Could not open message body file: '%s'\n",path);
 	}
 	/* And cap the tail with paired MIME boundary.. */
 	fprintf(errfp, "--%s--\n", boundarystr);
