@@ -89,6 +89,10 @@ int	global_wrkcnt = 0;
 int	mailq_Q_mode = 0;
 int	syncstart = 0;		/* while set, the thread subsystem shall start
 				   no childs! */
+int	hashlevels = 0;		/* How many levels of hashes are supported
+				   at the transport directory ?  (To speed
+				   up the directory processing, file opens,
+				   etc.. ) */
 extern int forkrate_limit;	/* How many forks per second ? */
 
 static int vtxprep_skip      = 0;
@@ -103,13 +107,16 @@ static int newents_limit = 200;
 extern memtypes stickymem;
 
 static struct ctlfile *schedule __((int fd, char *file, struct config_entry *ce, long ino));
-static struct ctlfile *vtxprep __((struct ctlfile *));
+static struct ctlfile *vtxprep __((struct ctlfile *, char *));
 static int  vtxmatch __((struct vertex *, struct config_entry *));
 static void link_in __((int flag, struct vertex *vp, u_char *s));
 static int  lockverify __((struct ctlfile *, char *));
 static int  globmatch   __((char *, char*));
 static void vtxdo   __((struct vertex *, struct config_entry *, char *));
 
+extern void  cfp_mksubdirs __((char *, char*));
+extern char *cfpdirname __((int));
+
 static RETSIGTYPE sig_exit   __((int sig));
 static RETSIGTYPE sig_alarm  __((int sig));
 static RETSIGTYPE sig_iot    __((int sig));
@@ -146,15 +153,21 @@ extern void prversion __((char *name));
 struct dirstatname {
 	struct stat st;
 	long ino;
+	time_t	not_before;
 	char name[1]; /* Allocate enough size */
 };
 struct dirqueue {
+	int	wrksum;
+	int	sorted;
 	int	wrkcount;
 	int	wrkspace;
+	int	wrkcount2; /* Back-pushed material a *2 -queue */
+	int	wrkspace2;
 	struct dirstatname **stats;
+	struct dirstatname **stats2;
 };
 
-static int dirqueuescan __((const char *dir, struct dirqueue *dq));
+static int dirqueuescan __((const char *dir,struct dirqueue *dq,int subdirs));
 static int syncweb __((struct dirqueue *dq, struct config_entry *ce));
 
 int global_maxkids = 1000;
@@ -223,6 +236,11 @@ struct ctlfile *cfp;
 }
 
 
+struct dirqueue dirqb;
+struct dirqueue *dirq = &dirqb;
+
+char **ArgvSave;
+
 int
 main(argc, argv)
 	int argc;
@@ -233,15 +251,15 @@ main(argc, argv)
 	int i, daemonflg, c, errflg, version, fd;
 	long offout, offerr;
 	struct config_entry *tp;
-	struct dirqueue dqb;
-	struct dirqueue *dq = &dqb;
 	int freeze = 0;
 
 	/* setlinebuf(stderr);  -- no need for this ? */
 
+	ArgvSave = argv;
+
 	mytime(&sched_starttime);
 
-	memset(&dqb,0,sizeof(dqb));
+	memset(&dirqb,0,sizeof(dirqb));
 	dirscan_mesh = sp_init();
 
 	if ((progname = strrchr(argv[0], '/')) == NULL)
@@ -263,7 +281,7 @@ main(argc, argv)
 	dlyverbose = 0;
 	syncstart = 0;
 	verbose = errflg = version = 0;
-	while ((c = getopt(argc, argv, "divE:f:Fl:L:N:P:q:QR:SVW")) != EOF)
+	while ((c = getopt(argc, argv, "divE:f:Fl:HL:N:P:q:QR:SVW")) != EOF)
 		switch (c) {
 		case 'f':	/* override default config file */
 			config = optarg;
@@ -283,6 +301,10 @@ main(argc, argv)
 			  exit(1);
 			}
 			break;
+		case 'H':
+			if (hashlevels < 2)
+				++hashlevels;
+			break;
 		case 'L':	/* override default log file */
 			log = optarg;
 			break;
@@ -332,7 +354,7 @@ main(argc, argv)
 		}
 	if (errflg) {
 	  fprintf(stderr,
-		  "Usage: %s [-disvV -f configfile -L logfile -P postoffice -Q rendezvous]\n",
+		  "Usage: %s [-dHisvV -f configfile -L logfile -P postoffice -Q rendezvous]\n",
 		  progname);
 	  exit(128+errflg);
 	}
@@ -414,8 +436,8 @@ main(argc, argv)
 	  }
 	  detach();		/* leave worldy matters behind */
 	  mytime(&now);
-	  printf("%s: scheduler daemon (%s)\n\tstarted at %s\n",
-		 progname, Version, (char *)rfc822date(&now));
+	  printf("%s: scheduler daemon (%s)\n\tpid %d started at %s\n",
+		 progname, Version, getpid(), (char *)rfc822date(&now));
 	}
 	/* Actually we want this to act as daemon,
 	   even when not in daemon mode.. */
@@ -463,10 +485,10 @@ main(argc, argv)
 
 	queryipcinit();
 
-	dirqueuescan(".", dq);
+	dirqueuescan(".", dirq, 1);
 
 	vtxprep_skip_lock = 0;
-	syncweb(dq, tp);
+	syncweb(dirq, tp);
 
 	if (dlyverbose) verbose = dlyverbose;
 
@@ -480,10 +502,11 @@ main(argc, argv)
 
 	if (syncstart) {
 	  int startcount;
-	  dirqueuescan(".", dq);
-	  startcount = dq->wrkcount;
-	  while (dq->wrkcount > 0 && !mustexit) {
-	    syncweb(dq, tp);
+	  dirqueuescan(".", dirq, 1);
+	  startcount = dirq->wrksum;
+	  while (dirq->wrksum > 0 && !mustexit) {
+	    if (syncweb(dirq, tp) < 0)
+	      break;
 	    queryipccheck();
 	  }
 	  fprintf(stderr,"Synchronous startup completed, messages: %d (%d skipped) recipients: %d\n",
@@ -500,11 +523,15 @@ main(argc, argv)
 	  canexit = 0;
 
 	  if (now >= next_dirscan) {
-	    /* Directory scan time for new jobs .. */
-	    if (dirqueuescan(".", dq) < 150) {
+	    /* Directory scan time for new jobs ..    */
+	    /* Do it recursively every now and then,
+	       so that if we forget some jobs, they will
+	       become relearned soon enough.          */
+	    if (dirqueuescan(".", dirq, 
+			     (now >= next_idlecleanup)) < 150) {
 	      /* If we have more things to scan, don't quit yet! */
 	      next_dirscan = now + sweepinterval;	/* 10 seconds interval */
-	      if (dq->wrkcount > 100)
+	      if (dirq->wrksum > 100)
 		next_dirscan = now + 60; /* Lots of work, pick new jobs
 					    less frequently */
 	    }
@@ -516,20 +543,22 @@ main(argc, argv)
 	  }
 
 	  /* Submit one item from pre-scheduler-queue into the scheduler */
-	  if (dq->wrkcount > 100) {
+	  if (dirq->wrksum > 100 ) {
 	    /* If more than 100 in queue, submit 20 in a burst.. */
 	    int i;
 	    for (i=0; i < 20; ++i)
-	      syncweb(dq, tp);
+	      syncweb(dirq, tp);
 	  }
-	  if (dq->wrkcount > 0)
-	    syncweb(dq, tp);
+	  if (dirq->wrksum > 0)
+	    syncweb(dirq, tp);
 
 	  /* See when to timeout from mux() */
 	  timeout = next_dirscan;
 
 	  /* If we still have things in pre-scheduler queue... */
-	  if (dq->wrkcount > 0)
+	  if (dirq->wrksum > 0 &&
+	      dirq->wrkcount > 0 &&
+	      dirq->stats[ dirq->wrkcount-1 ]->not_before <= now)
 	    timeout = now;
 
 	  /* Submit possible new jobs (unless frozen) */
@@ -551,7 +580,7 @@ main(argc, argv)
 	    }
 	    mytime(&now);
 	  } while (mux(timeout) == 0 && !gotalarm &&
-		   dq->wrkcount == 0 && !mustexit && timeout > now);
+		   dirq->wrksum == 0 && !mustexit && timeout > now);
 	} while (!mustexit);
 
 	/* Doing nicely we would kill the childs here, but we are not
@@ -609,15 +638,103 @@ int sig;
  * Absorb any new files that showed up in our directory.
  */
 
-static int in_dirscanqueue(dq,ino)
-	struct dirqueue *dq;
+int
+dq_insert(DQ, ino, file, delay)
+	void *DQ;
+	long ino;
+	char *file;
+	int delay;
+{
+	struct stat stbuf;
+	struct dirstatname *dsn;
+	struct dirqueue *dq = DQ;
+
+	mytime(&now);
+
+	if (dq == NULL)
+	  dq = dirq;
+
+	if (lstat(file,&stbuf) != 0 ||
+	    !S_ISREG(stbuf.st_mode)) {
+	  /* Not a regular file.. Let it be there for the manager
+	     to wonder.. */
+	  return -1;
+	}
+
+	/* 
+	   if (DQ == NULL)
+	   fprintf(stderr,"dq_insert(NULL,ino=%ld, file='%s', delay=%d)\n",
+	   ino,file,delay);
+	 */
+
+	/* Is it already in the database ? */
+	if (sp_lookup((u_long)ino,dirscan_mesh) != NULL) {
+	  fprintf(stderr,"scheduler: tried to dq_insert(ino=%ld, file='%s') already in queue\n",ino, file);
+	  return 1; /* It is! */
+	}
+
+	/* Now store the entry */
+	dsn = (struct dirstatname*)emalloc(sizeof(*dsn)+strlen(file));
+	memcpy(&(dsn->st),&stbuf,sizeof(stbuf));
+	dsn->ino = ino;
+	dsn->not_before = now + delay;
+	strcpy(dsn->name,file);
+
+	sp_install(ino,(u_char*)dsn,0,dirscan_mesh);
+
+	/* Put the new entry into the normal queue, unless there
+	   is stuff at the delayed queue, OR the delay is positive! */
+
+	if (delay <= 0 && dq->wrkcount2 == 0) {
+	  /* Into the normal queue */
+	  if (dq->wrkspace <= dq->wrkcount) {
+	    /* Increase the space */
+	    dq->wrkspace = dq->wrkspace ? dq->wrkspace << 1 : 8;
+	    if (dq->stats == NULL)
+	      dq->stats = (struct dirstatname **)emalloc(sizeof(void*) *
+							 dq->wrkspace);
+	    else
+	      dq->stats = (struct dirstatname**)erealloc(dq->stats,
+							 sizeof(void*) *
+							 dq->wrkspace);
+	  }
+	  dq->stats[dq->wrkcount] = dsn;
+	  dq->wrkcount += 1;
+	  dq->wrksum   += 1;
+	  dq->sorted    = 0;
+	} else {
+	  /* Into the DELAYED QUEUE */
+	  if (dq->wrkspace2 <= dq->wrkcount2) {
+	    /* Increase the space */
+	    dq->wrkspace2 = dq->wrkspace2 ? dq->wrkspace2 << 1 : 8;
+	    if (dq->stats2 == NULL)
+	      dq->stats2 = (struct dirstatname **)emalloc(sizeof(void*) *
+							  dq->wrkspace2);
+	    else
+	      dq->stats2 = (struct dirstatname**)erealloc(dq->stats2,
+							  sizeof(void*) *
+							  dq->wrkspace2);
+	  }
+	  dq->stats2[dq->wrkcount2] = dsn;
+	  dq->wrkcount2 += 1;
+	  dq->wrksum    += 1;
+	}
+	return 0;
+}
+
+int in_dirscanqueue(DQ,ino)
+	void *DQ;
 	long ino;
 {
 	int i;
+	struct dirqueue *dq = DQ;
+
+	if (dq == NULL)
+	  dq = dirq;
 
 	/* Return 1, if can find the "ino" in the queue */
 
-	if (dq->wrkcount == 0 || dirscan_mesh == NULL) return 0;
+	if (dq->wrksum == 0 || dirscan_mesh == NULL) return 0;
 	if (sp_lookup((u_long)ino,dirscan_mesh) != NULL) return 1;
 	return 0;
 }
@@ -627,20 +744,27 @@ const void *a; const void *b;
 {
 	const struct dirstatname **A = (const struct dirstatname **)a;
 	const struct dirstatname **B = (const struct dirstatname **)b;
+	int rc;
 
-	return ((*A)->st.st_ctime - (*B)->st.st_ctime);
+#if 0
+	if ((*A)->not_before > now) return  1;
+	if ((*B)->not_before > now) return -1;
+#endif
+
+	rc = ((*A)->st.st_ctime - (*B)->st.st_ctime);
+	return rc;
 }
 
-static int dirqueuescan(dir, dq)
+static int dirqueuescan(dir, dq, subdirs)
 	const char *dir;
 	struct dirqueue *dq;
+	int subdirs;
 {
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat stbuf;
 	char file[MAXNAMLEN+1];
 	int newents = 0;
-	struct dirstatname *dsn;
 
 #if 0
 	static time_t modtime = 0;
@@ -653,7 +777,9 @@ static int dirqueuescan(dir, dq)
 	modtime = stbuf.st_mtime;
 #endif
 
-if (verbose) { printf("dirqueuescan(dir='%s') ",dir); fflush(stdout); }
+	if (verbose && dir[0] == '.') {
+	  printf("dirqueuescan(dir='%s') ",dir); fflush(stdout);
+	}
 
 	/* Some changes lately, open the dir and read it */
 
@@ -662,7 +788,28 @@ if (verbose) { printf("dirqueuescan(dir='%s') ",dir); fflush(stdout); }
 	  /* Scan filenames into memory */
 
 	  if (!syncstart && newents > newents_limit)
-	    break; /* At most XXX per one go */
+	    break; /* At most NNN per one go */
+
+	  if (subdirs &&
+	      dp->d_name[0] >= 'A' &&
+	      dp->d_name[0] <= 'Z' &&
+	      dp->d_name[1] ==  0 ) {
+	    /* We do this recursively.. */
+	    if (dir[0] == '.' && dir[1] == 0)
+	      strcpy(file, dp->d_name);
+	    else
+	      sprintf(file, "%s/%s", dir, dp->d_name);
+
+	    if (lstat(file,&stbuf) != 0 ||
+		!S_ISDIR(stbuf.st_mode)) {
+	      /* Not a directory.. Let it be there for the manager
+		 to wonder.. */
+	      continue;
+	    }
+	    /* Recurse into levels below.. */
+	    newents += dirqueuescan(file, dq, subdirs);
+	    continue;
+	  } /* End of directories of names "A" .. "Z" */
 
 	  if (dp->d_name[0] >= '0' &&
 	      dp->d_name[0] <= '9') {
@@ -673,7 +820,10 @@ if (verbose) { printf("dirqueuescan(dir='%s') ",dir); fflush(stdout); }
 	      /* Already in pre-schedule-queue */
 	      continue;
 
-	    strcpy(file, dp->d_name);
+	    if (dir[0] == '.' && dir[1] == 0)
+	      strcpy(file, dp->d_name);
+	    else
+	      sprintf(file, "%s/%s", dir, dp->d_name);
 
 	    /* We may have this file in processing state...  */
 	    {
@@ -687,45 +837,12 @@ if (verbose) { printf("dirqueuescan(dir='%s') ",dir); fflush(stdout); }
 	      }
 	    }
 
-
-	    if (lstat(file,&stbuf) != 0 ||
-		!S_ISREG(stbuf.st_mode)) {
-	      /* Not a regular file.. Let it be there for the manager
-		 to wonder.. */
+	    if (dq_insert(dq, ino, file, -1))
 	      continue;
-	    }
 
-	    /* Now store the entry */
-	    dsn = (struct dirstatname*)emalloc(sizeof(*dsn)+strlen(file));
-	    memcpy(&(dsn->st),&stbuf,sizeof(stbuf));
-	    dsn->ino = ino;
-	    strcpy(dsn->name,file);
-	    sp_install(ino,(u_char*)dsn,0,dirscan_mesh);
-
-	    if (dq->wrkspace <= dq->wrkcount) {
-	      /* Increase the space */
-	      dq->wrkspace += 8;
-	      if (dq->stats == NULL)
-		dq->stats = (struct dirstatname **)emalloc(sizeof(void*) *
-							   dq->wrkspace);
-	      else
-		dq->stats = (struct dirstatname**)erealloc(dq->stats,
-							   sizeof(void*) *
-							   dq->wrkspace);
-	    }
-	    dq->stats[dq->wrkcount] = dsn;
-	    dq->wrkcount += 1;
 	    ++newents;
 	  } /* ... end of "filename starts with a [0-9]" */
 	}
-	if (newents && dq->wrkcount > 1) {
-	  /* Sort the dq->stats[] per file ctime -- LATEST on slot 0.
-	     (we drain this queue from the END) */
-	  qsort((void*)dq->stats,
-		dq->wrkcount,
-		sizeof(void*),
-		dq_ctimecompare );
-	}
 
 #ifdef	BUGGY_CLOSEDIR
 	/*
@@ -737,7 +854,8 @@ if (verbose) { printf("dirqueuescan(dir='%s') ",dir); fflush(stdout); }
 #endif
 	closedir(dirp);
 
-	if (verbose) printf("wrkcount=%d new=%d\n",dq->wrkcount,newents);
+	if (verbose && dir[0] == '.')
+	  printf("wrksum=%d new=%d\n",dq->wrksum,newents);
 
 	return newents;
 }
@@ -753,20 +871,52 @@ static int syncweb(dq, cehead)
 	long ino;
 
 	/* Any work to do ? */
-	if (dq->wrkcount == 0) return 0;
+	if (dq->wrksum == 0) return -1;
 
 	/* Be responsive, check query channel */
 	queryipccheck();
 
+	if (dq->wrkcount == 0 && dq->wrkcount2 > 0) {
+	  /* Ok, the primary queue is empty, move the delayed
+	     queue into the primary one. */
+	  if (dq->stats != NULL) free(dq->stats);
+	  dq->stats = dq->stats2;
+	  dq->stats2 = NULL;
+	  dq->wrkcount  = dq->wrkcount2;
+	  dq->wrkcount2 = 0;
+	  dq->wrkspace  = dq->wrkspace2;
+	  dq->wrkspace2 = 0;
+	  dq->sorted = 0;
+	}
+
+	if (!dq->sorted) {
+	  /* Sort the dq->stats[] per file ctime -- LATEST on slot 0.
+	     (we drain this queue from the END) */
+	  mytime(&now);
+	  qsort((void*)dq->stats,
+		dq->wrkcount,
+		sizeof(void*),
+		dq_ctimecompare );
+	  dq->sorted = 1;
+	}
+
+	mytime(&now);
+
+	if (dq->stats[dq->wrkcount-1]->not_before > now)
+	  return -1; /* WAIT! */
+
 	/* Ok some, decrement the count to change it to index */
 	dq->wrkcount -= 1;
+	dq->wrksum   -= 1;
 	file  =   dq->stats[dq->wrkcount]->name;
 	stbuf = &(dq->stats[dq->wrkcount]->st);
 	ino   =   dq->stats[dq->wrkcount]->ino;
 	/* Now we have pointers */
 
 	/* Deletion from the  dirscan_mesh  should ALWAYS succeed.. */
-	sp_delete(sp_lookup((u_long)ino,dirscan_mesh),dirscan_mesh);
+	spl = sp_lookup((u_long)ino,dirscan_mesh);
+	if (spl != NULL)
+	  sp_delete(spl,dirscan_mesh);
 
 	/* Sometimes we may get files already in processing
 	   into our pre-schedule queue */
@@ -803,23 +953,33 @@ void resync_file(file)
 	struct stat stbuf;
 	struct spblk *spl;
 	long ino, fd;
+	char *s;
 
 	lstat(file,&stbuf);
-	ino = atoi(file);
+
+	s = strrchr(file,'/');
+	if (s) ++s; else s = file;
+	ino = atoi(s);
 
 	/* Sometimes we may get reports of files
 	   already deleted from processing.. */
-	printf("Resyncing file \"%s\" (ino=%d)",file,ino);
 
 	/* Already in processing ? */
 	spl = sp_lookup((u_long)ino, spt_mesh[L_CTLFILE]);
 	if (spl == NULL) {
-	  printf(" .. not in processing database\n");
+	  if (!in_dirscanqueue(NULL,ino)) {
+	    printf("Resyncing file \"%s\" (ino=%d)",file,ino);
+	    printf(" .. not in processing database\n");
+	  }
 	  /* Not (anymore) in processing! */
 	  return;
 	}
+	printf("Resyncing file \"%s\" (ino=%d)",file,ino);
 	printf(" .. in processing db\n");
 
+	/* cfp_free()->unvertex()->unctlfile() will do reinsertion */
+	/* dq_insert(NULL,ino,file,31); */
+
 	/* Delete it... */
 	cfp_free((struct ctlfile *)spl->data, spl);
 	/* It will return into processing at next queue scan.. */
@@ -844,7 +1004,7 @@ static struct ctlfile *schedule(fd, file, cehead, ino)
 	struct vertex *vp;
 
 	/* read and process the control file */
-	if ((cfp = vtxprep(slurp(fd, ino))) == NULL) {
+	if ((cfp = vtxprep(slurp(fd, ino), file)) == NULL) {
 	  if (!vtxprep_skip) {	/* Unless skipped.. */
 	    eunlink(file);	/* everything here has been processed */
 	    if (verbose)
@@ -856,7 +1016,7 @@ static struct ctlfile *schedule(fd, file, cehead, ino)
 	}
 	if (cfp->head == NULL) {
 	  ++global_wrkcnt;
-	  unctlfile(cfp, 0);
+	  unctlfile(cfp, 0); /* Delete the file too */
 	  return NULL;
 	}
 	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
@@ -916,6 +1076,7 @@ slurp(fd, ino)
 	memset((void*)cfp, 0, sizeof(struct ctlfile));
 
 	cfp->fd = fd;
+	cfp->dirind = -1; /* Not known -- or top-level */
 	cfp->uid = stbuf.st_uid;
 	cfp->envctime = stbuf.st_ctime;
 	cfp->contents = contents;
@@ -1020,8 +1181,10 @@ lockverify(cfp,cp)		/* Return 1 when lock process does not exist */
 	int	lockpid;
 	int	sig = 0;
 
-#ifdef SIGCONT	/* OSF/1 acts differently if we use SIGNAL 0 :-( */
+#if 0
+# ifdef SIGCONT	/* OSF/1 acts differently if we use SIGNAL 0 :-( */
 	sig = SIGCONT;
+# endif
 #endif
 
 	++cp;
@@ -1048,8 +1211,9 @@ lockverify(cfp,cp)		/* Return 1 when lock process does not exist */
  *  wrapped into same vertex node with its respective ``recipient group''.
  *
  */
-static struct ctlfile *vtxprep(cfp)
+static struct ctlfile *vtxprep(cfp, file)
 	struct ctlfile *cfp;
+	char *file;
 {
 	register int i, opcnt;
 	register long *lp;
@@ -1063,10 +1227,15 @@ static struct ctlfile *vtxprep(cfp)
 	struct stat stbuf;
 	struct offsort *offarr;
 	int offspc;
+	long ino;
+
+	char fpath[128], path[128], path2[128];
 
 	if (cfp == NULL)
 	  return NULL;
 
+	strcpy(fpath, file);
+
 	channel = host = NULL;
 
 	/* copy offsets into local array */
@@ -1101,11 +1270,20 @@ static struct ctlfile *vtxprep(cfp)
 	      ++vtxprep_skip;
 	      ++vtxprep_skip_lock;
 	      free(offarr);
+
+	      /* XXX: Should we  dq_insert()  this back in ?
+		      Or do we let the occasional recursive
+		      scanner to check things deeply ? */
+	      cp = strrchr(file,'/');
+	      if (cp) ino = atol(cp+1); else ino = atol(file);
+	      dq_insert(NULL, ino, file, 32);
+
 	      return NULL;
 	    }
 	    *cp = _CFTAG_NORMAL; /* unlock it */
 	    lockaddr(cfp->fd, NULL, (long) (cp - cfp->contents),
-		     _CFTAG_LOCK, _CFTAG_NORMAL);
+		     _CFTAG_LOCK, _CFTAG_NORMAL, cfp->mid,
+		     cp+_CFTAG_RCPTPIDSIZE);
 	  }
 	  /* Calculate summary info */
 	  if (cp[-1] == _CF_RECIPIENT) {
@@ -1258,8 +1436,81 @@ static struct ctlfile *vtxprep(cfp)
 	  }
 	}
 	close(cfp->fd);	/* closes the fd opened in schedule() */
-	if (cfp->mid != NULL)
-	  sprintf(mfpath, "../%s/%s", QUEUEDIR, cfp->mid);
+
+	if (fpath[0] >= 'A') {
+	  /* Prefixed with a subdir path */
+	  int hash = 0;
+	  char *s = fpath;
+
+	  while (*s >= 'A' && *s <= 'Z') {
+	    hash <<= 8;
+	    hash |= (*s) & 0xff;
+	    ++s;
+	    if (*s != '/') break;
+	    ++s;
+	  }
+	  cfp->dirind = hash;
+
+	} else { /* Not at subdirs */
+
+	  if (hashlevels > 0) {
+	    /* We have a desire to use subdirs, now the magic of hashing
+	       into subdirs...  inode number ? */
+
+	    int hash;
+
+	    if (hashlevels > 1) {
+	      hash = atol(cfp->mid) % (26*26);
+	      hash = (('A' + hash / 26) << 8) | ('A' + hash % 26);
+	    } else {
+	      hash = (atol(cfp->mid) % 26) + 'A';
+	    }
+	    cfp->dirind = hash;
+	    /* Ok, we have the hash values, now move the file
+	       to match with our hashes.. */
+	    sprintf(path, "%s/%s", cfpdirname(hash), fpath);
+	    if (rename(fpath, path) != 0) {
+	      /* Failed, why ? */
+	      if (errno != ENOENT) {
+		/* For any other than 'no such (target) directory' */
+		cfp->dirind = -1;
+	      } else {
+		cfp_mksubdirs("",cfpdirname(hash));
+		if (rename(fpath, path) != 0)
+		  cfp->dirind = -1; /* Failed for any reason */
+		else
+		  strcpy(fpath, path);
+	      }
+	    } else
+		strcpy(fpath, path);
+
+	    if (cfp->dirind >= 0) {
+	      /* Successfully renamed the transport file to a subdir,
+		 now do the same to the queue directory! */
+	      sprintf(path,  "../%s/%s", QUEUEDIR, cfp->mid);
+	      sprintf(path2, "../%s/%s/%s",
+		      QUEUEDIR, cfpdirname(cfp->dirind), cfp->mid);
+
+	      if (rename(path,path2) != 0) {
+		if (errno == ENOENT) {
+		  /* No dirs ?? */
+		  cfp_mksubdirs(QUEUEDIR,cfpdirname(hash));
+		  rename(path, path2);
+		}
+	      }
+	      /* If failed, it will be reported below */
+	    }
+	  }
+	}
+
+
+	if (cfp->mid != NULL) {
+	  if (cfp->dirind >= 0)
+	    sprintf(mfpath, "../%s/%s/%s",
+		    QUEUEDIR, cfpdirname(cfp->dirind), cfp->mid);
+	  else
+	    sprintf(mfpath, "../%s/%s", QUEUEDIR, cfp->mid);
+	}
 	if (cfp->mid == NULL || cfp->logident == NULL
 	    || estat(mfpath, &stbuf) < 0) {
 	  if (cfp->vfp != NULL) {
@@ -1766,6 +2017,9 @@ static void init_timeserver()
 	}
 	if (ppid < 0) return; /* Error ?? brr.. */
 
+	ArgvSave[1] = NULL;
+	strcpy(ArgvSave[0],"TimeServer");
+
 	ppid = getppid(); /* who is our parent ? */
 
 	for(;;) {
@@ -1815,3 +2069,44 @@ time_t *timep;
 #endif
 	return time(timep); /* The classical old version.. */
 }
+
+char *
+cfpdirname(hash)
+int hash;
+{
+	static char dirbuf[8];
+
+	if (hash < 0)
+	  return ".";
+	if (hash < 256) {
+	  sprintf(dirbuf, "%c", hash);
+	} else {
+	  sprintf(dirbuf, "%c/%c", (hash >> 8) & 0xff, hash & 0xff);
+	}
+	return dirbuf;
+}
+
+
+/* We make one, or at most two subdirs */
+
+void
+cfp_mksubdirs(topdir,newpath)
+char *topdir, *newpath;
+{
+	char path[256];
+	int omask = umask(022);
+
+	if (*topdir != 0)
+	  sprintf(path, "../%s/%s", topdir, newpath);
+	else
+	  strcpy(path, newpath);
+
+	if (mkdir(path,0755) != 0) {
+	  char *s = strrchr(path,'/');
+	  if (s) *s = 0;
+	  mkdir(path,0755);
+	  if (s) *s = '/';
+	  mkdir(path,0755);
+	}
+	umask(omask);
+}
