@@ -227,16 +227,27 @@ unctlfile(cfp, no_unlink)
 	struct spblk *spl;
 	char	path[MAXPATHLEN+1];
 
+	if (cfp->dirind > 0)
+	  sprintf(path, "%s/%s", cfpdirname(cfp->dirind), cfp->mid);
+	else
+	  strcpy(path, cfp->mid);
+
 	if (!no_unlink) {
 	  reporterrs(cfp);
 #ifdef	LOG_INFO
 	  syslog(LOG_INFO, "%s: done", cfp->mid);
 #endif	/* LOG_INFO */
-	  eunlink(cfp->mid);
+	  eunlink(path);
 	  if (verbose)
-	    printf("%s: unlink %s (mid=0x%p)", cfp->logident, cfp->mid, cfp->mid);
+	    printf("%s: unlink %s (mid=0x%p)", cfp->logident, path, cfp->mid);
+
+	  if (cfp->dirind > 0)
+	    sprintf(path, "../%s/%s/%s",
+		    QUEUEDIR, cfpdirname(cfp->dirind), cfp->mid);
+	  else
+	    sprintf(path, "../%s/%s",
+		    QUEUEDIR, cfp->mid);
 
-	  sprintf(path, "../%s/%s", QUEUEDIR, cfp->mid);
 	  eunlink(path);
 	  if (verbose)
 	    printf("   and %s/\n", path);
@@ -247,6 +258,10 @@ unctlfile(cfp, no_unlink)
 		    "scheduler done processing %s\n", cfp->mid);
 	    fclose(cfp->vfp);
 	  }
+	} else {
+	  /* We will LOOSE this from the schedules -- add info about
+	     it into the indirscanqueue -- at the tail... */
+	  dq_insert(NULL, atol(cfp->mid), path, 30);
 	}
 	spl = sp_lookup(cfp->id, spt_mesh[L_CTLFILE]);
 	if (spl != NULL)
@@ -266,7 +281,10 @@ void unvertex(vp, justfree, ok)
 	if (vp->ngroup > 0)
 	  return;
 
-if (verbose && justfree < 0) fprintf(stderr,"unvertex(vtx=0x%p (thr=0x%p proc=0x%p, ng=%d) ,%d,%d)\n",vp,vp->thread,vp->proc,vp->ngroup,justfree,ok);
+	if (verbose && justfree < 0)
+	  fprintf(stderr,
+		  "unvertex(vtx=0x%p (thr=0x%p proc=0x%p, ng=%d) ,%d,%d)\n",
+		  vp,vp->thread,vp->proc,vp->ngroup,justfree,ok);
 
 	if (vp->thread->vertices == vp && vp->thread->proc &&
 	    vp->thread->proc->vertex == vp) {
@@ -275,6 +293,7 @@ if (verbose && justfree < 0) fprintf(stderr,"unvertex(vtx=0x%p (thr=0x%p proc=0x
 	    vp->proc = vp->thread->proc;
 	  vp->proc->vertex = vp;
 	}
+
 	if (vp->proc && vp->proc->vertex == vp) {
 	  vp->proc->fed = 1;     /* Mark it fed just in case.. */
 	  vp->proc->overfed = 0; /* .. and clear this .. */
@@ -284,7 +303,10 @@ if (verbose && justfree < 0) fprintf(stderr,"unvertex(vtx=0x%p (thr=0x%p proc=0x
 	    fprintf(stderr,
 		    "unvertex(vtx=0x%p,%d,%d) failed to pick_next_vertex()!\n",
 		    vp,justfree,ok);
-	    abort();
+	    /* We may become called with child feeder yet unflushed;
+	       shall we kill the kid ? (pick_next_vertex won't change
+	       vertex then..) */
+	    /* abort(); */
 	  }
 	}
 	if (vp->thread->proc && vp->thread->proc->vertex == vp) {
@@ -339,8 +361,10 @@ static struct vertex *findvertex(inum, offset, idx)
 	/* It is NOT POSSIBLE to cache cfp, based on the inum */
 	spl = sp_lookup((u_long)inum, spt_mesh[L_CTLFILE]);
 	if (spl == NULL || (cfp = (struct ctlfile *)spl->data) == NULL) {
-	  fprintf(stderr, "%s: cannot find control file for %ld!\n",
-		  progname, inum);
+	  /* It may have been kicked into input queue */
+	  if (!in_dirscanqueue(NULL,(long)inum))
+	    fprintf(stderr, "%s: cannot find control file for %ld!\n",
+		    progname, inum);
 	  return NULL;
 	}
 	for (i = 0; i < cfp->nlines; ++i) {
@@ -379,15 +403,27 @@ static int ctlowner(cfp)
 	char *path;
 	struct stat stbuf;
 	static int nope = -9;
+	int rc = nope;
 
 	if (cfp->mid == NULL)
 	  abort();
-	path = emalloc(5+strlen(cfp->mid)+sizeof QUEUEDIR);
-	sprintf(path, "../%s/%s", QUEUEDIR, cfp->mid);
-	if (stat(path, &stbuf) < 0)
-	  return --nope;
+#ifdef USE_ALLOCA
+	path = alloca(5+strlen(cfp->mid)+sizeof QUEUEDIR+8);
+#else
+	path = emalloc(5+strlen(cfp->mid)+sizeof QUEUEDIR+8);
+#endif
+	if (cfp->dirind > 0)
+	  sprintf(path, "../%s/%s/%s",
+		  QUEUEDIR, cfpdirname(cfp->dirind), cfp->mid);
+	else
+	  sprintf(path, "../%s/%s",
+		  QUEUEDIR, cfp->mid);
+	if (stat(path, &stbuf) == 0)
+	  rc = stbuf.st_uid;
+#ifndef USE_ALLOCA
 	free(path);
-	return stbuf.st_uid;
+#endif
+	return rc;
 }
 
 void
