@@ -48,8 +48,8 @@ char *Copyright2= "Copyright 1991-1996 Matti Aarnio";
 #include <sys/time.h>
 #include <sys/stat.h>
 #include <sys/file.h>
-#ifdef FCNTL_H
-# include FCNTL_H
+#ifdef HAVE_FCNTL_H
+# include <fcntl.h>
 #endif
 #include "zmsignal.h"
 #include <errno.h>
@@ -81,7 +81,7 @@ extern int wait();
 # include <sys/socket.h>
 # include <netinet/in.h>
 #endif
-#ifdef	HAVE_SYSLOG
+#ifdef	HAVE_SYSLOG_H
 # include <syslog.h>
 #endif	/* HAVE_SYSLOG */
 #include <identuser.h>
@@ -191,6 +191,16 @@ int	pid, routerpid;
 FILE	*logfp = NULL;
 int	D_alloc = 0;
 
+static char *m200 = "2.0.0";
+static char *m400 = "4.0.0";
+static char *m430 = "4.3.0";
+static char *m454 = "4.5.4";
+static char *m540 = "5.4.0";
+static char *m550 = "5.5.0";
+static char *m551 = "5.5.1";
+static char *m552 = "5.5.2";
+static char *m554 = "5.5.4";
+
 /*
  * The "style" variable controls when the router is interrogated about the
  * validity of something.  It is a string of letter-flags:
@@ -253,7 +263,7 @@ extern u_char *rfc822date __((time_t *));
 static int mvdata __((FILE *, char *));
 static int mvbdata __((FILE *, char *, long, int *));
 #if defined(HAVE_STDARG_H) && defined(HAVE_VPRINTF)
-extern void type __((int code, char *fmt, ...));
+extern void type __((int code, char *status, char *fmt, ...));
 #else
 extern void type __((/* int code, char *fmt, ... */));
 #endif
@@ -907,12 +917,12 @@ char *buf, *cp;
 		cp = NULL;
 		break;
 	  }
-	  type(503, cp);
+	  type(503, NULL, cp);
 	  return;
 	}
 	while (*cp == ' ' || *cp == '\t') ++cp;
 	if (sscanf(cp, "%s", helobuf) != 1) {
-	  type(501, "What is your domain name?");
+	  type(501, NULL, "What is your domain name?");
 	  return;
 	}
 	/*
@@ -928,10 +938,10 @@ char *buf, *cp;
 	 */
 	if (checkhelo && skeptical && partridge(helobuf, cfinfo)) {
 	  rfc821_error_ptr += (cp - helobuf);
-	  type821err(-501, buf,
+	  type821err(-501, "", buf,
 		     "Invalid `%s' parameter!",
 		     buf);
-	  type(501,"Err: %s", rfc821_error);
+	  type(501, "", "Err: %s", rfc821_error);
 	  strcpy(helobuf,"Bad.Helo.Input");
 	  return;
 	}
@@ -939,8 +949,8 @@ char *buf, *cp;
 	   complaining.. */
 	if (!checkhelo && partridge(helobuf, cfinfo)) {
 	  rfc821_error_ptr += (cp - helobuf);
-	  type821err(-250, buf, "Invalid `%s' parameter!", buf);
-	  type(-250,"Err: %s", rfc821_error);
+	  type821err(-250, "", buf, "Invalid `%s' parameter!", buf);
+	  type(-250, NULL, "Err: %s", rfc821_error);
 	  if (carp->cmd == Hello)
 	    strcpy(helobuf,"Bad.HELO.Input");
 	  else
@@ -949,9 +959,9 @@ char *buf, *cp;
 	cfinfo = findcf(helobuf);
 	if (cfinfo != NULL && *(cfinfo->flags) == '!') {
 	  if (cfinfo->flags[1] != '\0')
-	    type(501, "%s", (cfinfo->flags)+1);
+	    type(501, NULL, "%s", (cfinfo->flags)+1);
 	  else
-	    type(501, "Sorry, access denied.");
+	    type(501, NULL, "Sorry, access denied.");
 	  return;
 	}
 	/* Check helobuf corresponds to the reverse address */
@@ -959,17 +969,17 @@ char *buf, *cp;
 	    && rhostname[0] != '\0' && rhostname[0] != '['
 	    && !CISTREQ(helobuf, rhostname)) {
 	  if (checkhelo) {
-	    type(-250, "%s That hostname is inconsistent with",
+	    type(-250, NULL, "%s That hostname is inconsistent with",
 	         myhostname);
-	    type(-250, "%s your address to name mapping.",
+	    type(-250, NULL, "%s your address to name mapping.",
 	         myhostname);
 	  }
-	  type(carp->cmd==Hello2 ? -250: 250,
+	  type(carp->cmd==Hello2 ? -250: 250, NULL,
 	       "%s expected \"%s %s\"", myhostname,
 	     carp->verb, rhostname);
 	} else
-	  type(carp->cmd==Hello2 ? -250: 250,
-	       "%s Hello %s", myhostname, helobuf);
+	    type(carp->cmd==Hello2 ? -250: 250, NULL,
+		 "%s Hello %s", myhostname, helobuf);
 	if (carp->cmd == Hello2) {
 
 	  /* ESMTP -- RFC 1651,1652,1653,1428 thingies */
@@ -978,17 +988,18 @@ char *buf, *cp;
 	  sprintf(sizebuf,"SIZE %ld",maxsize); /* 0: No fixed max size
 						  in force, else:
 						  The FIXED maximum */
-	  type(-250,sizebuf);
-	  type(-250,"8BITMIME");
-	  type(-250,"PIPELINING");
-	  type(-250,"CHUNKING"); /* RFC 1830 */
-	  type(-250,"EXPN");
-	  type(-250,"VRFY");
-	  type(-250,"XDSN"); /* XX: Change to "DSN" when all parts of it function! */
-	  type(-250,"X-RCPTLIMIT 10000"); /* VERY HIGH figure, normal is 100 */
-	  type(-250,"X-TURNME");
-	  type(-250,"ETRN");
-	  type( 250,"HELP");
+	  type(-250,NULL, sizebuf);
+	  type(-250,NULL, "8BITMIME");
+	  type(-250,NULL, "PIPELINING");
+	  type(-250,NULL, "CHUNKING"); /* RFC 1830 */
+	  type(-250,NULL, "ENHANGEDSTATUSCODES");
+	  type(-250,NULL, "EXPN");
+	  type(-250,NULL, "VRFY");
+	  type(-250,NULL, "XDSN"); /* XX: Change to "DSN" when all parts of it function! */
+	  type(-250,NULL, "X-RCPTLIMIT 10000"); /* VERY HIGH figure, normal is 100 */
+	  type(-250,NULL, "X-TURNME");
+	  type(-250,NULL, "ETRN");
+	  type( 250,NULL, "HELP");
 	  with_protocol = WITH_ESMTP;
 	}
 	state = MailOrHello;
@@ -1014,7 +1025,7 @@ int insecure;
 
 	if (mfp == NULL
 	    && (mfp = mail_open(MSG_RFC822)) == NULL) {
-	  type(452, (char *)NULL);
+	  type(452,m430, "mail_open() failed with errno %d: %s", errno, strerror(errno));
 	  return;
 	}
 
@@ -1038,27 +1049,27 @@ int insecure;
 		cp = NULL;
 		break;
 	  }
-	  type(503, cp);
+	  type(503, m551, cp);
 	  return;
 	}
 	if (!CISTREQN(cp, "From:", 5)) {
-	  type(501, "where is From: in that?");
+	  type(501, m552, "where is From: in that?");
 	  return;
 	}
 	for (cp = cp+5; *cp != '\0' && *cp != '<'; ++cp)
 	  if (!isascii(*cp) || !isspace(*cp)) {
-	    type(501, "where is <...> in that?");
+	    type(501, m552, "where is <...> in that?");
 	    return;
 	  }
 	if (*cp == '\0') {
-	  type(501, "where is <...> in that?");
+	  type(501, m552, "where is <...> in that?");
 	  return;
 	} else if (*cp != '<') {
-	  type(501, "strangeness between : and <");
+	  type(501, m552, "strangeness between : and <");
 	  return;
 	}
 	if (*(cp+1) == '<') {
-	  type(501, "there are too many <'s in that!");
+	  type(501, m552, "there are too many <'s in that!");
 	  return;
 	}
 
@@ -1066,11 +1077,11 @@ int insecure;
 	s = rfc821_path(cp,STYLE(cfinfo, 'R'));
 	if (s == cp) {
 	  /* Failure.. */
-	  type821err(501,buf,"Path data: %s",rfc821_error);
+	  type821err(501,m552,buf,"Path data: %s",rfc821_error);
 	  return;
 	}
 	if (*s == '>') {
-	  type(501, "there are too many >'s in that!");
+	  type(501, m552, "there are too many >'s in that!");
 	  return;
 	}
 	++cp;		/* Skip the initial '<' */
@@ -1087,7 +1098,7 @@ int insecure;
 	    /* Actually we do not use this data... */
 	    s += 5;
 	    if (bodytype != NULL) {
-	      type(501,"BODY= double definition!");
+	      type(501,m554,"BODY= double definition!");
 	      rc = 1;
 	      break;
 	    }
@@ -1102,7 +1113,7 @@ int insecure;
 	      s += 4;
 	    }
 	    if (*s && *s != ' ' && *s != '\t') {
-	      type(455,"BODY-param data error, must be one of: 8BITMIME/BINARYMIME/7BIT");
+	      type(455, m454, "BODY-param data error, must be one of: 8BITMIME/BINARYMIME/7BIT");
 	      rc = 1;
 	      break;
 	    }
@@ -1111,7 +1122,7 @@ int insecure;
 	  if (CISTREQN("SIZE=",s,5)) {
 	    s += 5;
 	    if (sizeoptval != -1) {
-	      type(455,"SIZE-param double definition!");
+	      type(455,m554,"SIZE-param double definition!");
 	      rc = 1;
 	      break;
 	    }
@@ -1123,7 +1134,7 @@ int insecure;
 	      ++s;
 	    }
 	    if (*s && *s != ' ' && *s != '\t') {
-	      type(501,"SIZE-param data error");
+	      type(501,m454,"SIZE-param data error");
 	      rc = 1;
 	      break;
 	    }
@@ -1132,14 +1143,14 @@ int insecure;
 	  /* IETF-NOTARY  SMTP-DRPT extensions */
 	  if (CISTREQN("ENVID=",s,6)) {
 	    if (drpt_envid != NULL) {
-	      type(501,"ENVID double definition!");
+	      type(501,m554,"ENVID double definition!");
 	      rc = 1;
 	      break;
 	    }
 	    drpt_envid = s+6;
 	    while (*s && *s != ' ' && *s != '\t') {
 	      if (*s < 33 || *s > 126) {
-		type(501,"ENVID data contains illegal characters!");
+		type(501,m554,"ENVID data contains illegal characters!");
 		rc = 1;
 		break;
 	      }
@@ -1147,13 +1158,13 @@ int insecure;
 	    }
 	    *s++ = 0;
 	    if (strlen(drpt_envid) < 6) {
-	      type(501,"ENVID= missing data!");
+	      type(501,m454,"ENVID= missing data!");
 	      rc = 1;
 	      break;
 	    }
 	    continue;
 	  }
-	  type(555,"Unknown MAIL FROM:<> parameter: %s",s);
+	  type(555,m454,"Unknown MAIL FROM:<> parameter: %s",s);
 	  break;
 	}
 	if (rc != 0) return; /* Error(s) in previous loop.. */
@@ -1177,7 +1188,7 @@ int insecure;
 	    (mfp = mail_open(MSG_RFC822)) == NULL) {
 	  if (s)
 	    free(s);
-	  type(452, (char *)NULL);
+	  type(452,m430, (char *)NULL);
 	  return;
 	}
 	rewind(mfp);
@@ -1203,17 +1214,17 @@ int insecure;
 	sender_ok = 0; /* Set it, when we are sure.. */
 
 	if (ferror(mfp)) {
-	  type(452, (char *)NULL);
+	  type(452,m430,(char *)NULL);
 	  clearerr(mfp);
 	} else if (sizeoptval > maxsize && maxsize > 0) {
-	  type(552, "The message size exceeds fixed maximum message size");
+	  type(552,"5.3.4","The message size exceeds fixed maximum message size");
 	} else if (sizeoptval > availspace) {
-	  type(452, "Try again later, insufficient storage available at the moment");
+	  type(452,"4.3.1","Try again later, insufficient storage available at the moment");
 	} else {
 	  if (s)
 	    type(atoi(s), s+4, "Ok", "Ok");
 	  else
-	    type(250, (char *)NULL);
+	    type(250, "2.1.0", (char *)NULL);
 	  sender_ok = 1;
 	}
 	if (s)
@@ -1242,27 +1253,27 @@ char *buf, *cp;
 		cp = NULL;
 		break;
 	  }
-	  type(503, cp);
+	  type(503, m551, cp);
 	  return;
 	}
 	if (!CISTREQN(cp, "To:", 3)) {
-	  type(501, "where is To: in that?");
+	  type(501, m552, "where is To: in that?");
 	  return;
 	}
 	for (cp = cp+3; *cp != '\0' && *cp != '<'; ++cp)
 	  if (!isascii(*cp) || !isspace(*cp)) {
-	    type(501, "where is <...> in that?");
+	    type(501, m552, "where is <...> in that?");
 	    return;
 	  }
 	if (*cp == '\0') {
-	  type(501, "where is <...> in that?");
+	  type(501, m552, "where is <...> in that?");
 	  return;
 	} else if (*cp != '<') {
-	  type(501, "strangeness between : and <");
+	  type(501, m552, "strangeness between : and <");
 	  return;
 	}
 	if (*(cp+1) == '<') {
-	  type(501, "there are too many <'s in that!");
+	  type(501, m552, "there are too many <'s in that!");
 	  return;
 	}
 
@@ -1270,11 +1281,11 @@ char *buf, *cp;
 	s = rfc821_path(cp,STYLE(cfinfo, 'R'));
 	if (s == cp) {
 	  /* Failure.. */
-	  type821err(501,buf,"Path data: %s",rfc821_error);
+	  type821err(501,"5.5.2",buf,"Path data: %s",rfc821_error);
 	  return;
 	}
 	if (*s == '>') {
-	  type(501, "there are too many >'s in that!");
+	  type(501, m552, "there are too many >'s in that!");
 	  return;
 	}
 	++cp;		/* Skip the initial '<' */
@@ -1283,7 +1294,7 @@ char *buf, *cp;
 	/*printf("  <path>: len=%d \"%s\"\n",cp-s,cp);*/
 
 	if (*cp == '\0') {
-	  type(501, "what is a null recipient?");
+	  type(501, "5.1.3", "what is a null recipient?");
 	  return;
 	}
 
@@ -1296,7 +1307,7 @@ char *buf, *cp;
 	  /* IETF-NOTARY  SMTP-RCPT-DRPT extensions */
 	  if (CISTREQN("NOTIFY=",s,7)) {
 	    if (drpt_notify) {
-	      type(501,"NOTIFY-param double defined!");
+	      type(501,m554,"NOTIFY-param double defined!");
 	      return;
 	    }
 	    drpt_notify = s;
@@ -1310,7 +1321,7 @@ char *buf, *cp;
 	    else if (CISTREQN("NEVER",s,5))
 	      s += 5;
 	    if (*s && *s != ' ' && *s != '\t') {
-	      type(455,"NOTIFY-param data error");
+	      type(455,m454,"NOTIFY-param data error");
 	      return;
 	    }
 	    if (*s) *s++ = '\0';
@@ -1318,7 +1329,7 @@ char *buf, *cp;
 	  }
 	  if (CISTREQN("RET=",s,4)) {
 	    if (drpt_ret) {
-	      type(501,"RET-param double defined!");
+	      type(501,m554,"RET-param double defined!");
 	      return;
 	    }
 	    drpt_ret = s;
@@ -1328,7 +1339,7 @@ char *buf, *cp;
 	    else if (CISTREQN("NO",s,2))
 	      s += 2;
 	    if (*s && *s != ' ' && *s != '\t') {
-	      type(501,"RET-param data error");
+	      type(501,m454,"RET-param data error");
 	      return;
 	    }
 	    if (*s) *s++ = '\0';
@@ -1336,7 +1347,7 @@ char *buf, *cp;
 	  }
 	  if (CISTREQN("ORCPT=",s,6)) {
 	    if (drpt_orcpt) {
-	      type(501,"ORCPT-param double defined!");
+	      type(501,m554,"ORCPT-param double defined!");
 	      return;
 	    }
 	    drpt_orcpt = s;
@@ -1344,7 +1355,7 @@ char *buf, *cp;
 	      ++s;
 	    if (*s) *s++ = '\0';
 	    if (strlen(drpt_orcpt) < 7) {
-	      type(501,"ORCPT-param data error!");
+	      type(501,m454,"ORCPT-param data error!");
 	      return;
 	    }
 	    continue;
@@ -1381,20 +1392,20 @@ char *buf, *cp;
 	  sizeoptval = 0;
 	sizeoptsum += sizeoptval;
 	if (ferror(mfp)) {
-	  type(452, (char *)NULL);
+	  type(452,m430,(char *)NULL);
 	  clearerr(mfp);
 	} else if (maxsize > 0 && sizeoptsum > maxsize) {
-	  type(552, "Message size exceeds fixed maximum size for acceptable email");
+	  type(552, "5.3.4", "Message size exceeds fixed maximum size for acceptable email");
 	} else if (sizeoptsum > availspace) {
-	  type(452, "insufficient storage space, try again later");
+	  type(452, "4.3.1", "insufficient storage space, try again later");
 	} else if (s) {
 	  type(atoi(s), s+4, "Ok", "Ok");
 	  ++rcpt_count;
 	} else {
 	  if (sizeoptval)
-	    type(250, "Ok; can accomodate %d byte message",sizeoptval);
+	    type(250, "2.1.5", "Ok; can accomodate %d byte message",sizeoptval);
 	  else
-	    type(250, (char *)NULL);
+	    type(250, "2.1.5", (char *)NULL);
 	  ++rcpt_count;
 	}
 	if (s)
@@ -1408,7 +1419,7 @@ char *name, *cp;
 {
 	FILE *mfp = mail_open(MSG_RFC822);
 	if (!mfp) {
-	  type(452, "Failed to initiate TURNME request;  Disk full?");
+	  type(452,m400,"Failed to initiate TURNME request;  Disk full?");
 	  fflush(stdout);
 	  return;
 	}
@@ -1416,11 +1427,11 @@ char *name, *cp;
 	/* printf("050-My uid=%d/%d\r\n",getuid(),geteuid()); */
 	runasrootuser();
 	if (mail_close_scheduler(mfp)) {
-	  type(452, "Failed to initiate TURNME request;  Permission denied?");
+	  type(452,m400,"Failed to initiate TURNME request;  Permission denied?");
 	} else {
-	  type(-250,"A TURNME request is initiated - lets hope the system");
-	  type(-250,"has resources to honour it.   We call the remote, if");
-	  type( 250,"we have anything to send there.");
+	  type(-250,m200,"A TURNME request is initiated - lets hope the system");
+	  type(-250,m200,"has resources to honour it.   We call the remote, if");
+	  type( 250,m200,"we have anything to send there.");
 	}
 	runastrusteduser();
 	fflush(stdout);
@@ -1454,25 +1465,25 @@ smtp_data(buf,cp)
 		cp = NULL;
 		break;
 	  }
-	  type(503, cp);
+	  type(503, m552, cp);
 	  fflush(stdout);
 	  return 0;
 	}
 	if (sender_ok == 0) {
-	  type(550, "No valid sender, rejecting all recipients");
+	  type(550, "5.1.7", "No valid sender, rejecting all recipients");
 	  fflush(stdout);
 	  state = MailOrHello;
 	  return 0;
 	}
 	if (rcpt_count == 0) {
 	  /* No valid recipients! */
-	  type(550, "No valid recipient at RCPT addresses, or no RCPT addresses at all");
+	  type(550, "5.1.3", "No valid recipient at RCPT addresses, or no RCPT addresses at all");
 	  fflush(stdout);
 	  state = MailOrHello;
 	  return 0;
 	}
 	rcpt_count = 0; /* now we can zero it.. */
-	type(354, (char *)NULL);
+	type(354, NULL, (char *)NULL);
 	fflush(stdout);
 	fputs("env-end\n",mfp);
 	/* We set alarm()s inside the mvdata() */
@@ -1489,7 +1500,7 @@ smtp_data(buf,cp)
 	if (*msg != 0) {
 	  mail_abort(mfp);
 	  mfp = NULL;
-	  type(452,"%s",msg);
+	  type(452,m430,"%s",msg);
 	  fflush(stdout);
 	} else if (s_feof()) {
 	  /* [mea@utu.fi] says this can happen */
@@ -1499,7 +1510,7 @@ smtp_data(buf,cp)
 	  fflush(stdout);
 	  return -1;
 	} else if (ferror(mfp)) {
-	  type(452, (char *)NULL);
+	  type(452,m430,(char *)NULL);
 	  fflush(stdout);
 	  clearerr(mfp);
 	  mail_abort(mfp);
@@ -1508,16 +1519,16 @@ smtp_data(buf,cp)
 	} else if (maxsize > 0 && filsiz > maxsize) {
 	  mail_abort(mfp);
 	  mfp = NULL;
-	  type(552, "Message size exceeded fixed maximum size for acceptable email");
+	  type(552, "5.3.4", "Message size exceeded fixed maximum size for acceptable email");
 	  fflush(stdout);
 	} else if (mail_close(mfp) == EOF) {
-	  type(452, (char *)NULL);
+	  type(452,m430,(char *)NULL);
 	  fflush(stdout);
 	  mfp = NULL;
 	  reporterr(tell, "message file close failed");
 	} else {
 	  mfp = NULL;
-	  type(250, "Roger");
+	  type(250, "2.6.0", "Roger");
 	  fflush(stdout);
 #ifdef	LOG_INFO
 	  syslog(LOG_INFO,
@@ -1561,7 +1572,7 @@ smtp_bdata(buf,cp)
 	bdata_last = CISTREQ(msg,"LAST");
 	if (!(bdata_chunksize > 0L
 	      && (rc == 1 || (rc == 2 && bdata_last)))) {
-	  type(501,NULL);
+	  type(501,m552,NULL);
 	  return 0;
 	}
 	
@@ -1594,7 +1605,7 @@ smtp_bdata(buf,cp)
 		cp = NULL;
 		break;
 	  }
-	  type(503, cp);
+	  type(503, m552,cp);
 	  fflush(stdout);
 	  if (mfp) mail_abort(mfp);
 	  mfp = NULL;
@@ -1606,7 +1617,7 @@ smtp_bdata(buf,cp)
 	    char *cp = "No valid sender, rejecting all recipients";
 	    if (sender_ok != 0)
 	      cp = "No valid recipient at RCPT addresses, or no RCPT addresses at all";
-	    type(550, cp);
+	    type(550, "5.1.3", cp);
 	    fflush(stdout);
 	    state = MailOrHello;
 	    if (mfp) mail_abort(mfp);
@@ -1619,7 +1630,7 @@ smtp_bdata(buf,cp)
 	}
 
 	if (mfp == NULL) {
-	  type(503,"BDAT block discarded due to earlier error");
+	  type(503,m552,"BDAT block discarded due to earlier error");
 	} else if (*msg != 0) {
 	  mail_abort(mfp);
 	  clearerr(mfp);
@@ -1633,7 +1644,7 @@ smtp_bdata(buf,cp)
 	  reporterr(tell, "premature EOF on BDAT input");
 	  return -1;
 	} else if (ferror(mfp)) {
-	  type(452, (char *)NULL);
+	  type(452,m400,(char *)NULL);
 	  clearerr(mfp);
 	  mail_abort(mfp);
 	  clearerr(mfp);
@@ -1642,19 +1653,19 @@ smtp_bdata(buf,cp)
 	} else if (maxsize > 0 && tell > maxsize) {
 	  mail_abort(mfp);
 	  mfp = NULL;
-	  type(552, "Message size exceeded fixed maximum size for acceptable email");
+	  type(552, "5.3.4", "Message size exceeded fixed maximum size for acceptable email");
 	} else if (bdata_last) {
 	  if (fstat(FILENO(mfp), &stbuf) < 0)
 	    inum = 0;
 	  else
 	    inum = stbuf.st_ino;
 	  if (mail_close(mfp) == EOF) {
-	    type(452, (char *)NULL);
+	    type(452,m400,(char *)NULL);
 	    mfp = NULL;
 	    reporterr(tell, "message file close failed");
 	  } else {
 	    mfp = NULL;
-	    type(250, "Roger, stored %ld bytes into spool", tell);
+	    type(250, "2.6.0", "Roger, stored %ld bytes into spool", tell);
 #ifdef	LOG_INFO
 	    syslog(LOG_INFO,
 		   "accepted id %d (%ldc) from %s/%d",
@@ -1666,7 +1677,7 @@ smtp_bdata(buf,cp)
 	    }
 	  }
 	} else { /* Not last chunk! */
-	  type(250,"Received %ld bytes",bdata_chunksize);
+	  type(250,"2.6.0","Received %ld bytes",bdata_chunksize);
 	}
 	if (bdata_last) {
 	  state = MailOrHello;
@@ -1732,7 +1743,7 @@ smtpserver(insecure)
 	if (wantconn(s_inputfd, "smtp-receiver") == 0) {
 	  syslog(LOG_WARNING, "refusing connection from %s/%s",
 		 rhostname, ident_username);
-	  type(421, "%s ZMailer Server %s WILL NOT TALK TO YOU at %s",
+	  type(421, NULL, "%s ZMailer Server %s WILL NOT TALK TO YOU at %s",
 	       myhostname, VersionNumb, cp);
 	  fflush(stdout);
 	  sleep(1);
@@ -1740,7 +1751,7 @@ smtpserver(insecure)
 	}
 #endif
 
-	type(220, "%s ZMailer Server %s ESMTP%s ready at %s",
+	type(220, NULL, "%s ZMailer Server %s ESMTP%s ready at %s",
 	     myhostname, VersionNumb, ident_flag ? "+IDENT":"", cp);
 	fflush(stdout);
 
@@ -1802,7 +1813,7 @@ smtpserver(insecure)
 	    break;
 	  }
 	  if (c != '\n') {
-	    type(500, "Line too long/not terminated with CRLF..");
+	    type(500,m552, "Line too long/not terminated with CRLF..");
 	    continue;
 	  }
 
@@ -1822,7 +1833,7 @@ smtpserver(insecure)
 
 	  if (rc >= 0) {
 	    rfc821_error_ptr = buf + rc;
-	    type821err(500, buf,
+	    type821err(500, m552, buf,
 		       "Illegal input characters: %s",
 		       ((buf[rc]=='\0') ? "NUL on SMTP input" :
 			((buf[rc]&0x80) ? "8-bit char on SMTP input":
@@ -1838,7 +1849,7 @@ smtpserver(insecure)
 	  for (cp = buf; isascii(*cp) && isalpha(*cp); ++cp)
 	    continue;
 	  if (cp >  buf + 6) {	/* "DEBUG" is longest of them.. */
-	    type(550, "Syntax error");
+	    type(550,m552,"Syntax error");
 	    fflush(stdout);
 	    continue;
 	  }
@@ -1851,7 +1862,7 @@ smtpserver(insecure)
 	  }
 	  *cp++ = c;
 	  if (carp->verb == NULL) {
-	    type(550, "Unknown command '%s'", buf);
+	    type(550, m552, "Unknown command '%s'", buf);
 #ifdef	LOG_WARNING
 	    syslog(LOG_WARNING,
 		   "unknown SMTP command '%s' from %s/%d",
@@ -1862,7 +1873,7 @@ smtpserver(insecure)
 	  }
 	  switch (carp->cmd) {
 	    case Null:
-		type(550, "panic!");
+		type(550,m550,"panic!");
 		fflush(stdout);
 		break;
 	    case Hello:
@@ -1914,7 +1925,7 @@ smtpserver(insecure)
 		}
 		if (state != Hello)
 		  state = MailOrHello;
-		type(250, (char *)NULL);
+		type(250,m200,NULL);
 		fflush(stdout);
 		break;
 	    case Help:
@@ -1924,11 +1935,11 @@ smtpserver(insecure)
 	    case Verify:
 	    case Verify2:
 		if (state == Hello) {
-		  type (503, "Waiting for HELO/EHLO command");
+		  type(503, m551, "Waiting for HELO/EHLO command");
 		  fflush(stdout);
 		  break;
 		}
-		if ((cfi = STYLE(cfinfo, 'v'))) {
+		{
 		  char *s;
 		  while (*cp == ' ' || *cp == '\t') ++cp;
 		  if (*cp == '<')
@@ -1936,26 +1947,30 @@ smtpserver(insecure)
 		  else
 		    s = rfc821_path2(cp,cfi); /* Without < > */
 		  if (s == cp) {
-		    type821err(501,buf,"Path data: %s",rfc821_error);
+		    type821err(501,m552, buf,"Path data: %s",rfc821_error);
 		    break;
 		  }
 		  while (*s == ' ' || *s == '\t') ++s;
 		  if (*s != 0) {
-		    type(501, "Growl! Extra junk after the VRFY argument!");
+		    type(501, m552, "Growl! Extra junk after the VRFY argument!");
 		    break;
 		  }
+		}
+		if ((cfi = STYLE(cfinfo, 'v'))) {
+		  char *s;
 		  if ((s = router(RKEY_VERIFY, 0, cp)) != NULL) {
 		    printf("%s\r\n", s);
 		    free(s);
 		  } else
-		    type(501, "Unable to verify that address");
+		    type(501,m540, "Unable to verify that address");
 		} else
-		  type(252, (char *)NULL);
+		  type(252, "2.5.2", (char *)NULL); /* Syntax ok */
+
 		fflush(stdout);
 		break;
 	    case Expand:
 		if (state == Hello) {
-		  type (503, "Waiting for HELO/EHLO command");
+		  type(503, m551, "Waiting for HELO/EHLO command");
 		  fflush(stdout);
 		  break;
 		}
@@ -1967,21 +1982,21 @@ smtpserver(insecure)
 		  else
 		    s = rfc821_path2(cp,cfi); /* Without < > */
 		  if (s == cp) {
-		    type821err(501,buf,"Path data: %s",rfc821_error);
+		    type821err(501,m552,buf,"Path data: %s",rfc821_error);
 		    break;
 		  }
 		  while (*s == ' ' || *s == '\t') ++s;
 		  if (*s != 0) {
-		    type(501, "Growl! Extra junk after the EXPN argument!");
+		    type(501, m552, "Growl! Extra junk after the EXPN argument!");
 		    break;
 		  }
 		  if ((s = router(RKEY_EXPAND, 0, cp)) != NULL) {
 		    printf("%s\r\n", s);
 		    free(s);
 		  } else
-		    type(501, "Unable to verify that address");
+		    type(501,m540, "Unable to verify that address");
 		} else
-		  type(502, (char *)NULL);
+		  type(502,m540, (char *)NULL);
 		fflush(stdout);
 		break;
 	    case Turnme:
@@ -1989,17 +2004,17 @@ smtpserver(insecure)
 		fflush(stdout);
 		break;
 	    case Turn:
-		type(502, (char *)NULL);
+		type(502, m551, (char *)NULL);
 		fflush(stdout);
 		break;
 	    case NoOp:
-		type(250, (char *)NULL);
+		type(250, m200,  (char *)NULL);
 		fflush(stdout);
 		break;
 	    case Verbose:
-		type(-250, VerbID, Version);
-		type(-250, Copyright);
-		type(250, Copyright2);
+		type(-250, m200, VerbID, Version);
+		type(-250, m200, Copyright);
+		type(250,  m200, Copyright2);
 		fflush(stdout);
 		VerboseCommand = 1;
 		break;
@@ -2008,14 +2023,14 @@ smtpserver(insecure)
 		fflush(stdout);
 		break;
 	    case Tick:
-		type(250, "%s", buf);
+		type(250, m200, "%s", buf);
 		fflush(stdout);
 		with_protocol = WITH_BSMTP;
 		break;
 	    case Quit:
 		if (mfp != NULL)
 		  mail_abort(mfp);
-		type(221, (char *)NULL, "Out");
+		type(221, m200, NULL, "Out");
 		fflush(stdout);
 		return;
 	    default:
@@ -2027,8 +2042,8 @@ smtpserver(insecure)
 	reporterr(lseek(0,0,SEEK_CUR), "session terminated");
 }
 
-#define	TYPE_(m)	type(-214, "%s", m);
-#define	TYPE(m)		type(214, "%s", m);
+#define	TYPE_(m)	type(-214, NULL, "%s", m);
+#define	TYPE(m)		type(214, NULL, "%s", m);
 
 /*
  * parse the query string and print an appropriate help message.
@@ -2109,7 +2124,7 @@ help(cfinfo, query)
 			cp = "Unknown";
 			break;
 		}
-		type(214,"The current state is: %s.", cp);
+		type(214,NULL,"The current state is: %s.", cp);
 		break;
 	case Send:
 	case Send2:
@@ -2120,7 +2135,7 @@ help(cfinfo, query)
 		TYPE ("\tThis command will never be implemented.");
 		break;
 	case Turnme:
-		type(-214,"%s hostname", carp->verb);
+		type(-214,NULL,"%s hostname", carp->verb);
 		TYPE_("\tThis command schedules (at least tries to) all");
 		TYPE_("\toutbound traffic to ``hostname'' host.");
 		TYPE_("\tFor security reasons this server will initiate the");
@@ -2199,9 +2214,9 @@ help(cfinfo, query)
 		TYPE_("This mailer will always accept 8-bit and binary message data");
 		TYPE_("though you are better to use MIME format!");
 		TYPE_("");
-		type(-214,"For local information contact: postmaster@%s",
+		type(-214,NULL,"For local information contact: postmaster@%s",
 		     myhostname);
-		type(214,"SMTP server comments and bug reports to: <zmhacks@nic.funet.fi>");
+		type(214,NULL,"SMTP server comments and bug reports to: <zmhacks@nic.funet.fi>");
 		break;
 	}
 	fflush(stdout);
@@ -2450,12 +2465,12 @@ router(function, holdlast, args)
 	int sawend;
 
 	if (args == NULL) {
-	  type(501, (char *)NULL);
+	  type(501,NULL,NULL);
 	  return NULL;
 	}
 
 	if (routerpid <= 0 && (routerpid = callr()) <= 0) {
-	  type(451, (char *)NULL);
+	  type(451,NULL,NULL);
 	  return NULL;
 	}
 
@@ -2504,7 +2519,7 @@ router(function, holdlast, args)
 #ifdef HAVE_VPRINTF
 #ifdef HAVE_STDARG_H
 void
-type(int code, char *fmt, ...)
+type(int code, char *status, char *fmt, ...)
 #else
 /* VARARGS2 */
 void
@@ -2513,12 +2528,12 @@ type(code, fmt, va_alist)
 	char *fmt;
 	va_dcl
 #endif
-#else
+#else /* No VPRINTF */
 /* VARARGS2 */
 void
-type(code, fmt, s1, s2, s3, s4, s5, s6)
+type(code, status, fmt, s1, s2, s3, s4, s5, s6)
 	int code;
-	char *fmt, *s1, *s2, *s3, *s4, *s5, *s6;
+	char *status, *fmt, *s1, *s2, *s3, *s4, *s5, *s6;
 #endif
 {
 	char format[BUFSIZ];
@@ -2532,8 +2547,14 @@ type(code, fmt, s1, s2, s3, s4, s5, s6)
 		c = ' ';
 
 	printf("%03d%c", code, c);
-	if (logfp != NULL)
-		fprintf(logfp, "%dw\t%03d%c", pid, code, c);
+	if (status && status[0] != 0)
+	  printf("%s ", status);
+
+	if (logfp != NULL) {
+	  fprintf(logfp, "%dw\t%03d%c", pid, code, c);
+	  if (status && status[0] != 0)
+	    fprintf(logfp, "%s ", status);
+	}
 
 	switch (code) {
 	case 211: /* System status */
@@ -2648,16 +2669,16 @@ type(code, fmt, s1, s2, s3, s4, s5, s6)
 
 
 void
-type821err(code,inbuf,msg,s1,s2,s3)
+type821err(code,status,inbuf,msg,s1,s2,s3)
 	int code;
-	char *msg, *inbuf;
+	char *status, *msg, *inbuf;
 	void *s1, *s2, *s3; /* XX: Propably not portable.. */
 {
-	char *s = inbuf+4;
+	char *s = inbuf+3+strlen(status)+1;
 
-	printf("%03d-", code < 0 ? -code : code);
+	printf("%03d-%s ", code < 0 ? -code : code,status);
 	if (logfp != NULL)
-	  fprintf(logfp, "%dw\t%03d-", pid, code);
+	  fprintf(logfp, "%dw\t%03d-%s ", pid, code,status);
 	if (strlen(inbuf)>3)
 	  while (*s && s < rfc821_error_ptr) {
 	    ++s;
@@ -2668,7 +2689,7 @@ type821err(code,inbuf,msg,s1,s2,s3)
 	printf("^\n");
 	if (logfp != NULL)
 	  fprintf(logfp,"^\n");
-	type(code,msg,s1,s2,s3);
+	type(code,status,msg,s1,s2,s3);
 }
 
 static void header_to_mime __((char *, int *, int));
