@@ -1,5 +1,5 @@
 #! /bin/sh
-realprog=/usr/local/lib/vacation
+realprog=/l/mail/bin/vacation
 case "$HOME" in
 '')	echo "$0: HOME environment variable not set. Something is wrong." >&2
 	exit 1;;
@@ -16,7 +16,7 @@ start)
 		echo "the default one for you -- edit it if you want to"
 		echo "say when you return, or tell people whom to contact"
 		echo "in your absence"
-		cp /usr/local/share/vacation.msg $HOME/.vacation.msg
+		cp /l/mail/vacation.msg $HOME/.vacation.msg
 	else
 		echo "$HOME/.vacation.msg already exists.  Using it."
 	fi
@@ -28,7 +28,7 @@ start)
 	if test -f $HOME/.vacforward; then
 		echo $HOME/.vacforward exists.  Using it.
 	else
-		echo "$user," '"|/usr/local/lib/vacation' $user'"' > \
+		echo "$user," '"|/l/mail/bin/vacation' $user'"' > \
 			$HOME/.vacforward
 	fi
 	ln $HOME/.vacforward $HOME/.forward
