@@ -33,8 +33,8 @@
 #include <sys/stat.h>
 #include <sys/file.h>
 #include <sys/wait.h>
-#ifdef FCNTL_H
-# include FCNTL_H
+#ifdef HAVE_FCNTL_H
+# include <fcntl.h>
 #endif
 #include <sysexits.h>
 #include "mail.h"
@@ -48,6 +48,10 @@ extern int cistrcmp();
 extern char *routermxes();
 extern int errno;
 
+#ifndef strrchr
+extern char *strrchr();
+#endif
+
 static struct address *ctladdr __((char *cp));
 
 
@@ -215,6 +219,7 @@ ctlopen(file, channel, host, exitflagp, selectaddr, matchrouter)
 	int headers_cnt = 0;
 	int headers_spc = 0;
 	int largest_headersize = 80; /* Some magic minimum.. */
+	char dirprefix[8];
 
 	static struct ctldesc d; /* ONLY ONE OPEN AT THE TIME! */
 
@@ -222,6 +227,16 @@ ctlopen(file, channel, host, exitflagp, selectaddr, matchrouter)
 	  ctlsticky(NULL,NULL); /* Reset the internal state.. */
 
 	memset(&d,0,sizeof(d));
+	if (*file >= 'A') {
+	  /* Has some hash subdirectory in front of itself */
+	  strncpy(dirprefix,file,sizeof(dirprefix));
+	  dirprefix[sizeof(dirprefix)-1] = 0;
+	  s = strrchr(dirprefix,'/');
+	  if (s) *++s = 0;
+	  /*  "A/B/"  */
+	} else
+	    dirprefix[0] = 0;
+
 
 	d.msgfd = -1; /* Always the zero is not good for your health .. */
 	if ((d.ctlfd = open(file, O_RDWR, 0)) < 0) {
@@ -305,7 +320,11 @@ ctlopen(file, channel, host, exitflagp, selectaddr, matchrouter)
 	  return NULL;
 	}
 
-	d.ctlid = atol(file);
+	s = strrchr(file,'/');	/* In case the file in in a subdir.. */
+	if (s)
+	  d.ctlid = atol(s+1);
+	else
+	  d.ctlid = atol(file);
 	d.senders = NULL;
 	d.recipients = NULL;
 	d.rcpnts_total = 0;
@@ -367,7 +386,7 @@ ctlopen(file, channel, host, exitflagp, selectaddr, matchrouter)
 		|| (ap->routermxes == NULL && selectaddr == NULL
 		    && host != NULL && cistrcmp(host,ap->host) !=0)
 		|| !lockaddr(d.ctlfd, d.ctlmap, d.offset[i]+1,
-			     _CFTAG_NORMAL, _CFTAG_LOCK)) {
+			     _CFTAG_NORMAL, _CFTAG_LOCK, file, host)) {
 	      if (ap->routermxes)
 		free((char *)ap->routermxes);
 	      ap->routermxes = NULL;
@@ -378,7 +397,7 @@ ctlopen(file, channel, host, exitflagp, selectaddr, matchrouter)
 	    rp = (struct rcpt *)emalloc(sizeof (struct rcpt));
 	    if (rp == NULL) {
 	      lockaddr(d.ctlfd, d.ctlmap, d.offset[i]+1,
-		       _CFTAG_LOCK, _CFTAG_DEFER);
+		       _CFTAG_LOCK, _CFTAG_DEFER, file, host);
 	      warning("Out of virtual memory!", (char *)NULL);
 	      *exitflagp = 1;
 	      break;
@@ -538,11 +557,13 @@ ctlopen(file, channel, host, exitflagp, selectaddr, matchrouter)
 	}
 
 #ifdef USE_ALLOCA
-	mfpath = alloca((u_int)5+sizeof(QUEUEDIR)+strlen(d.msgfile));
+	mfpath = alloca((u_int)5 + sizeof(QUEUEDIR)
+			+ strlen(dirprefix) + strlen(d.msgfile));
 #else
-	mfpath = emalloc((u_int)5+sizeof(QUEUEDIR)+strlen(d.msgfile));
+	mfpath = emalloc((u_int)5 + sizeof(QUEUEDIR)
+			 + strlen(dirprefix) + strlen(d.msgfile));
 #endif
-	sprintf(mfpath, "../%s/%s", QUEUEDIR, d.msgfile);
+	sprintf(mfpath, "../%s/%s%s", QUEUEDIR, dirprefix, d.msgfile);
 	if ((d.msgfd = open(mfpath, O_RDONLY, 0)) < 0) {
 	  int e = errno;
 	  for (rp = d.recipients; rp != NULL; rp = rp->next) {
