@@ -206,7 +206,8 @@ diagnostic(rp, rc, timeout, fmt, va_alist) /* (rp, rc, timeout, "fmtstr", remote
 		printf(" %s", message);
 	printf("\n");
 	if (!lockaddr(rp->desc->ctlfd, rp->desc->ctlmap,
-		      rp->lockoffset, _CFTAG_LOCK, mark)) {
+		      rp->lockoffset, _CFTAG_LOCK, mark,
+		      rp->desc->msgfile, "?host?")) {
 	  /* something went wrong in unlocking it, concurrency problem? */
 	}
 	rp->lockoffset = 0;	/* mark this recipient unlocked */
