@@ -1,7 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
- *	Copyright 1992-1994 Matti Aarnio -- MIME processing et.al.
+ *	Copyright 1992-1996 Matti Aarnio -- MIME processing et.al.
  */
 
 /* History:
@@ -13,6 +13,22 @@
  * Still more modifications by Matti Aarnio <mea@nic.funet.fi>
  */
 
+/* ZENV variables used by this program:
+ *
+ *	ZCONFIG		-- passed to programs run at pipes
+ *	MAILBOX		-- same
+ *	MAILBIN		-- same
+ *	MAILSHARE	-- same
+ *	PATH		-- used to find programs, passed to programs
+ *	MBOXLOCKS	-- defines file/mbox lock mechanisms; a string with:
+ *		'.' -- "dotlock" mechanism for mail spool files
+ *		':' -- separates mailbox, and file lock schemes
+ *		'L' -- lockf() lock
+ *		'F' -- flock() lock
+ *		'N' -- NFSMBOX (uses special locking daemon at remote system)
+ */
+
+
 #include "hostenv.h"
 #include <stdio.h>
 #include <ctype.h>
@@ -101,6 +117,16 @@ struct utimbuf {
 #define	F_OK	0
 #endif	/* !F_OK */
 
+#ifdef HAVE_LOCKF
+char *MBOXLOCKS_default = ".L:L";
+#else
+#ifdef HAVE_FLOCK
+char *MBOXLOCKS_default = ".F:F";
+#else
+char *MBOXLOCKS_default = ".:"; /* Huh ?? No lockf() nor flock() ??? */
+#endif
+#endif
+
 #ifdef HAVE_DIRENT_H
 # include <dirent.h>
 #else /* not HAVE_DIRENT_H */
@@ -593,6 +619,151 @@ propably_x400(addr)
 	return 0;
 }
 
+/* Various lock acquisition routines */
+
+int havedotlock = 0, havemaillock = 0;
+
+int
+acquire_mboxlock(rp,file)
+struct rcpt *rp;
+char *file;
+{
+#ifdef	HAVE_MAILLOCK_H
+	char *maillockuser;
+	struct passwd *pw;
+	int i;
+	char errbuf[30];
+
+	if (*(rp->addr->user) == TO_FILE) {
+	  if ((pw = getpwuid(uid)) == NULL) {
+	    notaryreport(file,"failed",
+			 "5.2.1 (Target file has no known owner)",
+			 "510 (Target file has no known owner)");
+	    DIAGNOSTIC(rp, EX_NOPERM,
+		       "File \"%s\" has no owner", file);
+	    return 1;
+	  }
+	  maillockuser = pw->pw_name;
+	}
+	else
+	  maillockuser = rp->addr->user;
+
+	switch ((i = maillock(maillockuser, 2))) {
+	case L_SUCCESS:
+	  cp = NULL;
+	  break;
+	case L_MAXTRYS:
+	  notaryreport(file,"delayed",
+		       "4.4.5 (Mailbox unlocking fails)",
+		       "450 (Mailbox unlocking fails)");
+	  DIAGNOSTIC(rp, EX_TEMPFAIL, "\"%s\" is locked", file);
+	  return 1;
+	case L_NAMELEN:
+	  cp = "recipient name > 13 chars";
+	  break;
+	case L_TMPLOCK:
+	  cp = "problem creating temp lockfile";
+	  break;
+	case L_TMPWRITE:
+	  cp = "problem writing pid into temp lockfile";
+	  break;
+	case L_ERROR:
+	  cp = "Something other than EEXIST happened";
+	  break;
+	case L_MANLOCK:
+	  cp = "cannot set mandatory lock on temp lockfile";
+	  break;
+	default:
+	  sprintf(errbuf, "maillock() error %d", i);
+	  cp = errbuf;
+	  break;
+	}
+	if (cp != NULL) {
+	  DIAGNOSTIC3(rp, EX_UNAVAILABLE,
+		      "Error maillock()ing \"%s\": %s", file, cp);
+	  return 1;
+	}
+	havemaillock = 1;
+#endif	/* HAVE_MAILLOCK_H */
+#ifdef  HAVE_DOTLOCK
+	havedotlock = (dotlock(file) == 0);
+	if (!havedotlock) {
+	  char mbuf[256];
+	  sprintf(mbuf, "\"%s\": %s", file, strerror(errno));
+	  notaryreport(file,"delayed",
+		       "4.4.5 (File locking with dotlock failed)",
+		       "450 (File locking with dotlock failed)");
+	  DIAGNOSTIC(rp, errno == EBUSY ? EX_TEMPFAIL :
+		     (errno == EACCES ? EX_NOPERM : EX_UNAVAILABLE),
+		     "can't dotlock %s", mbuf);
+	  return 1;
+	}
+#endif  /* HAVE_DOTLOCK */
+	return 0;
+}
+
+int
+acquire_nfsmboxlock(rp,file)
+struct rcpt *rp;
+char *file;
+{
+#ifdef	USE_NFSMBOX
+	if (nfslock(file, LOCK_EX) != 0) {
+	  alarm(0);
+	  notaryreport(file,"delayed",
+		       "4.4.5 (File locking with NFS lock failed)",
+		       "450 (File locking with NFS lock failed)");
+	  DIAGNOSTIC(rp, EX_TEMPFAIL,
+		     "nfs lock() can't nfslock \"%s\"", file);
+	  return 1;
+	}
+#endif	/* USE_NFSMBOX */
+	return 0;
+}
+
+
+int
+acquire_lockflock(fdmail,rp,file)
+int fdmail;
+struct rcpt *rp;
+char *file;
+{
+#if defined(HAVE_LOCKF) && defined(F_LOCK)
+	/* Seek to begining for locking! [mea@utu.fi] */
+	if (lseek(fdmail, 0L, L_SET) < 0 ||
+	    lockf(fdmail, F_LOCK, 0) < 0) {
+	  alarm(0);
+	  notaryreport(file,"delayed",
+		       "4.4.5 (File locking with  lockf  failed)",
+		       "450 (File locking with  lockf  failed)");
+	  DIAGNOSTIC(rp, EX_TEMPFAIL,
+		     "can't lockf() \"%s\"", file);
+	  return 1;
+	}
+	lseek(fdmail, 0L, L_XTND); /* To the end of the file */
+#endif	/* HAVE_LOCKF */
+	return 0;
+}
+
+int
+acquire_flocklock(fdmail,rp,file)
+int fdmail;
+struct rcpt *rp;
+char *file;
+{
+#ifdef	HAVE_FLOCK
+	if (flock(fdmail, LOCK_EX) < 0) {
+	  alarm(0);
+	  notaryreport(file,"delayed",
+		       "4.4.5 (File locking with  flock  failed)",
+		       "450 (File locking with  flock  failed)");
+	  DIAGNOSTIC(rp, EX_TEMPFAIL,
+		     "can't flock() \"%s\"", file);
+	  return 1;
+	}
+#endif	/* HAVE_FLOCK */
+	return 0;
+}
 
 /*
  * deliver - deliver the letter in to user's mail box.  Return
@@ -612,8 +783,7 @@ deliver(dp, rp, timestring)
 	int hasdir;
 	struct stat st, s2;
 	char *file = NULL, *cp, *at, *plus;
-	int havedotlock = 0, havemaillock = 0;
-	int ismbox = 0;
+	int ismbox = 0, i;
 #if	defined(HAVE_SOCKET)
 	struct biffer *nbp = NULL;
 #ifdef	HAVE_PROTOCOLS_RWHOD_H
@@ -621,9 +791,17 @@ deliver(dp, rp, timestring)
 #endif
 #endif
 	struct passwd *pw;
-#ifdef	HAVE_MAILLOCK_H
-	char *maillockuser;
-#endif
+	char *mboxlocks = getzenv("MBOXLOCKS");
+	char *filelocks = NULL;
+	char *locks;
+
+	if (mboxlocks == NULL || *mboxlocks == 0 )
+	  mboxlocks = MBOXLOCKS_default;
+
+	if ((filelocks = strchr(mboxlocks,':')))
+	  ++filelocks;
+	else
+	  filelocks = "";
 
 	uid  = atol(rp->addr->misc);
 	at   = strchr(rp->addr->user,'@');
@@ -666,14 +844,18 @@ deliver(dp, rp, timestring)
 	    return;
 	  }
 	  file = strdup(rp->addr->user);
+	  
 	  if (access(file, F_OK) < 0) {
-	      fdmail = createfile(rp, file, uid);
-	      if (fdmail < 0) {
-		  free(file);
-		  setrootuid(rp);
-		  return;
-	      }
-	      close(fdmail);
+	    fdmail = createfile(rp, file, uid);
+	    /* Did the create fail for some other reason, than
+	       that the file exists already ? (Due to a race in
+	       between two processes creating the files.) */
+	    if (fdmail < 0 && errno != EEXIST) {
+	      free(file);
+	      setrootuid(rp);
+	      return;
+	    }
+	    close(fdmail);
 	  }
 	  setrootuid(rp);
 	  break;
@@ -891,144 +1073,95 @@ deliver(dp, rp, timestring)
 	if (!S_ISREG(st.st_mode))
 	  /* don't lock non-files */;
 	else {
+
+	  int err = 0;
+
+
+	  if (!ismbox)	/* Not mailbox, use file-locks */
+	    mboxlocks = filelocks;
+
+	  locks = mboxlocks;
+
+	  if (verboselog)
+	    fprintf(verboselog,"Locking sequence: '%s'\n",locks);
+
+	  while (*locks != 0) {
+	    switch(*locks) {
+	      case '"': /* ZENV variable may have quotes with it.. */
+		break;
+	      case ':':
+		err = -1; /* negative error is no REAL error */
+		break;
+	      case '.':
+		if (ismbox)
+		  err = acquire_mboxlock(rp,file);
+		break;
+	      case 'N':
+	      case 'n':
+		err = acquire_nfsmboxlock(rp,file);
+		break;
+	      case 'L':
+	      case 'l':
+		err = acquire_lockflock(fdmail,rp,file);
+		break;
+	      case 'F':
+	      case 'f':
+		err = acquire_flocklock(fdmail,rp,file);
+		break;
+	      default:
+		err = 1;
+		break;
+	    }
+	    ++locks; /* Always advance */
+	    if (err) break;
+	  }
+
+	  if (err > 0) {
+	    --locks;
+	    while (locks >= mboxlocks) {
+	      switch (*locks) {
+		case '"':
+		  break;
+		case '.':
 #ifdef	HAVE_MAILLOCK_H
-	  if (*(rp->addr->user) == TO_FILE) {
-	    if ((pw = getpwuid(uid)) == NULL) {
-	      notaryreport(file,"failed",
-			   "5.2.1 (Target file has no known owner)",
-			   "510 (Target file has no known owner)");
-	      DIAGNOSTIC(rp, EX_NOPERM,
-			 "File \"%s\" has no owner", file);
-	      return;
-	    }
-	    maillockuser = pw->pw_name;
-	  }
-	  else
-	    maillockuser = rp->addr->user;
-
-	  if (ismbox) {
-	    int i;
-	    char errbuf[30];
-
-	    switch ((i = maillock(maillockuser, 2))) {
-	    case L_SUCCESS:
-	      cp = NULL;
-	      break;
-	    case L_MAXTRYS:
-	      notaryreport(file,"delayed",
-			   "4.4.5 (Mailbox unlocking fails)",
-			   "450 (Mailbox unlocking fails)");
-	      DIAGNOSTIC(rp, EX_TEMPFAIL, "\"%s\" is locked", file);
-	      return;
-	    case L_NAMELEN:
-	      cp = "recipient name > 13 chars";
-	      break;
-	    case L_TMPLOCK:
-	      cp = "problem creating temp lockfile";
-	      break;
-	    case L_TMPWRITE:
-	      cp = "problem writing pid into temp lockfile";
-	      break;
-	    case L_ERROR:
-	      cp = "Something other than EEXIST happened";
-	      break;
-	    case L_MANLOCK:
-	      cp = "cannot set mandatory lock on temp lockfile";
-	      break;
-	    default:
-	      sprintf(errbuf, "maillock() error %d", i);
-	      cp = errbuf;
-	      break;
-	    }
-	    if (cp != NULL) {
-	      DIAGNOSTIC3(rp, EX_UNAVAILABLE,
-			  "Error locking \"%s\": %s", file, cp);
-	      return;
-	    }
-	    havemaillock = 1;
-	  }
+		  if (havemaillock && ismbox)
+		    mailunlock();
+		  havemaillock = 0;
 #endif	/* HAVE_MAILLOCK_H */
-#ifdef  HAVE_DOTLOCK
-	  if (ismbox &&
-	      !(havedotlock = dotlock(file) == 0)) {
-	    char mbuf[256];
-	    sprintf(mbuf, "\"%s\", errno %d", file, errno);
-	    DIAGNOSTIC(rp, errno == EBUSY ? EX_TEMPFAIL :
-		       (errno == EACCES ? EX_NOPERM : EX_UNAVAILABLE),
-		       "can't lock %s", mbuf);
-#ifdef	HAVE_MAILLOCK_H
-	    if (havemaillock)
-	      mailunlock();
-#endif
-	    return;
-	  }
-#endif  /* HAVE_DOTLOCK */
+#ifdef	HAVE_DOTLOCK
+		  if (ismbox)
+		    dotunlock(file);
+		  havedotlock = 0;
+#endif /*HAVE_DOTLOCK*/
+		  break;
+	        case 'N':
+		case 'n':
 #ifdef	USE_NFSMBOX
-	  if (nfslock(file, LOCK_EX) != 0) {
-	    alarm(0);
-#ifdef	HAVE_MAILLOCK_H
-	    if (havemaillock)
-	      mailunlock();
-#endif
-#ifdef  HAVE_DOTLOCK
-	    if (havedotlock)
-	      dotunlock(file);
-#endif /* HAVE_DOTLOCK */
-	    notaryreport(file,"delayed",
-			 "4.4.5 (File locking with NFS lock failed)",
-			 "450 (File locking with NFS lock failed)");
-	    DIAGNOSTIC(rp, EX_TEMPFAIL,
-		       "nfs lock() can't lock \"%s\"", file);
-	    setrootuid(rp);
-	    close(fdmail);
-	    return;
-	  }
+		  unlock(file);
 #endif	/* USE_NFSMBOX */
-#if defined(HAVE_LOCKF) && defined(F_LOCK)
-	  /* Seek to begining for locking! [mea@utu.fi] */
-	  if (lseek(fdmail,0L,L_SET) < 0 ||
-	      lockf(fdmail, F_LOCK, 0) < 0) {
-	    alarm(0);
-#ifdef	HAVE_MAILLOCK_H
-	    if (havemaillock)
-	      mailunlock();
+		  break;
+	        case 'L':
+		case 'l':
+#if defined(HAVE_LOCKF) && defined(F_LOCK) /* If one, also the other */
+		  lseek(fdmail,0L,L_SET);
+		  lockf(fdmail, F_ULOCK, 0);
+#endif	/* HAVE_LOCKF */
+		  break;
+	        case 'F':
+		case 'f':
+#ifdef HAVE_FLOCK
+		  flock(fdmail, LOCK_UN);
 #endif
-#ifdef  HAVE_DOTLOCK
-	    if (havedotlock)
-	      dotunlock(file);
-#endif /* HAVE_DOTLOCK */
-	    notaryreport(file,"delayed",
-			 "4.4.5 (File locking with  lockf  failed)",
-			 "450 (File locking with  lockf  failed)");
-	    DIAGNOSTIC(rp, EX_TEMPFAIL,
-		       "lockf() can't lock \"%s\"", file);
-	    setrootuid(rp);
+		  break;
+		default:
+		  break;
+	      }
+	      --locks;
+	    }
 	    close(fdmail);
+	    setrootuid(rp);
 	    return;
 	  }
-
-#endif	/* HAVE_LOCKF */
-#ifdef	HAVE_FLOCK
-	  if (flock(fdmail, LOCK_EX) < 0) {
-	    alarm(0);
-#ifdef	HAVE_MAILLOCK_H
-	    if (havemaillock)
-	      mailunlock();
-#endif
-#ifdef  HAVE_DOTLOCK
-	    if (havedotlock)
-	      dotunlock(file);
-#endif /* HAVE_DOTLOCK */
-	    notaryreport(file,"delayed",
-			 "4.4.5 (File locking with  flock  failed)",
-			 "450 (File locking with  flock  failed)");
-	    DIAGNOSTIC(rp, EX_TEMPFAIL,
-		       "flock() can't lock \"%s\"", file);
-	    setrootuid(rp);
-	    close(fdmail);
-	    return;
-	  }
-#endif	/* HAVE_LOCKF */
 	}
 
 	/* Turn off the alarm */
@@ -1041,31 +1174,65 @@ deliver(dp, rp, timestring)
 	if (nbp != NULL)
 	  nbp->offset = eofindex;
 #endif	/* BIFF || RBIFF */
+
 	fp = putmail(dp, rp, fdmail, "a+", timestring, file);
+	if (!fp)
+	  close(fdmail); /* Close now, as putmail() didn't
+			    do it at error return */
 	
 	if (S_ISREG(st.st_mode)) {
-#ifdef	USE_NFSMBOX
-	  unlock(file);
-#endif	/* USE_NFSMBOX */
-#if defined(HAVE_LOCKF) && defined(F_ULOCK)
-	  lseek(fdmail,0L,L_SET);
-	  lockf(fdmail, F_ULOCK, 0);/* XX: does this really work?*/
-	  /* Yes it does, but it needs seek to begining of
-	     the file at first! [mea@utu.fi] */
-#endif	/* HAVE_LOCKF */
-#if	defined(HAVE_FLOCK) || (!defined(USE_NFSMBOX) && !defined(HAVE_LOCKF) && !defined(HAVE_MAILLOCK_H))
-	  flock(fdmail, LOCK_UN);
-#endif	/* HAVE_FLOCK */
+
+	  /* Previously acquired locks need to be released,
+	     preferrably in reverse order */
+
+	  --locks;
+
+	  while (locks >= mboxlocks) {
+	    /* if (verboselog)
+	       fprintf(verboselog, "Unlock char: '%c'\n",*locks); */
+	    switch (*locks) {
+	    case '"':
+	      break;
+	    case '.':
 #ifdef	HAVE_MAILLOCK_H
-	  if (havemaillock)
-	    mailunlock();
+	      if (ismbox && havemaillock)
+		mailunlock();
+	      havemaillock = 0;
 #endif	/* HAVE_MAILLOCK_H */
 #ifdef	HAVE_DOTLOCK
-	  if (havedotlock)
-	    dotunlock(file);
+	      if (ismbox)
+		dotunlock(file);
+	      havedotlock = 0;
 #endif /*HAVE_DOTLOCK*/
+	      break;
+	    case 'N':
+	    case 'n':
+#ifdef	USE_NFSMBOX
+	      unlock(file);
+#endif	/* USE_NFSMBOX */
+	      break;
+	    case 'L':
+	    case 'l':
+#if defined(HAVE_LOCKF) && defined(F_LOCK)
+	      lseek(fdmail,0L,L_SET);
+	      lockf(fdmail, F_ULOCK, 0);
+#endif	/* HAVE_LOCKF */
+	      break;
+	    case 'F':
+	    case 'f':
+#ifdef HAVE_FLOCK
+	      flock(fdmail, LOCK_UN);
+#endif
+	      break;
+	    default:
+	      break;
+	    }
+	    --locks;
+	  }
 	}
+
 	setrootuid(rp);
+
 	if (fp != NULL) {
 	  fclose(fp);		/* this closes fdmail */
 	  notaryreport(rp->addr->user,"delivery",
@@ -1464,20 +1631,25 @@ creatembox(rp, filep, uid, gid, pw)
 #ifdef	HAVE_UTIMES
 	      struct timeval tv[2];
 	      stat(*filep,&st); /* This by all propability will not fail.. */
-	      tv[0].tv_sec = 1; /* Not zero, but close.. */
+	      tv[0].tv_sec = 0; /* never read */
 	      tv[1].tv_sec = st.st_mtime;
 	      tv[0].tv_usec = tv[1].tv_usec = 0;
 	      utimes(*filep, tv);
 #else
 	      struct utimbuf tv;
 	      stat(*filep,&st); /* This by all propability will not fail.. */
-	      tv.actime  = 1;	/* Not zero, but close.. */
+	      tv.actime  = 0;	/* never read */
 	      tv.modtime = st.st_mtime;
 	      utime(*filep, &tv);
 #endif
 	    }
 	    return 1;
 	  }
+	  if (errno == EEXIST) { /* It exists -- propably a race between
+				    two file creators caused this */
+	    
+	    return 1;
+	  }
 	}
 
 	/* assert *filep != NULL */
@@ -1504,6 +1676,8 @@ createfile(rp, file, uid)
 
 	if ((fd = open(file, O_RDWR|O_CREAT|O_EXCL, MAILMODE)) < 0) {
 	  saverrno = errno;
+	  if (errno == EEXIST)
+	    return -3;
 	  if ((cp = strrchr(file, '/')) != NULL) {
 	    *cp = '\0';
 	    if (exstat(rp, file, &st, stat) < 0) {
@@ -1532,6 +1706,9 @@ createfile(rp, file, uid)
 	  if ((fd = open(file, O_RDWR|O_CREAT|O_EXCL, MAILMODE)) < 0) {
 	    saverrno = errno;
 	    setrootuid(rp);
+	    errno = saverrno;
+	    if (errno == EEXIST)
+	      return -3;
 	  } else {
 	    setrootuid(rp);
 	    return fd;
