@@ -218,7 +218,12 @@ long ehlo_sizeval      = 0;
 int  esmtp_on_banner = -1;
 int  main_esmtp_on_banner = -1;
 int  servport = -1;
-
+int  smtp_bufsize  = 8*1024; /* Size of the buffer; this should be large
+				enough to contain MOST cases of pipelined
+				SMTP information, AND still fit within
+				roundtrip TCP buffers */
+int  smtp_outcount = 0;      /* we used this much.. */
+int  block_written = 0;      /* written anything in async phase */
 time_t now;
 
 extern int errno;
@@ -465,6 +470,9 @@ main(argc, argv, envp)
 	} else
 	  need_host = 1;
 
+	setvbuf(stdout, NULL, _IOFBF, 8096*4 /* 32k */);
+	fd_blockingmode(fileno(stdout)); /* Just to make sure.. */
+
 	if (myhostname[0] == 0) {
 	  /* Default it only when not having an explicite value
 	     in it..   James S MacKinnon <jmack@Phys.UAlberta.Ca> */
@@ -480,7 +488,7 @@ main(argc, argv, envp)
 
 	if (logfile != NULL) {
 	  if ((fd = open(logfile, O_CREAT|O_APPEND|O_WRONLY, 0644)) < 0)
-	    fprintf(stderr,
+	    fprintf(stdout,
 		    "#%s: cannot open logfile \"%s\"!\n",
 		    argv[0], logfile);
 	  else
@@ -840,12 +848,38 @@ deliver(dp, smtpfp, verboselog, startrp, endrp)
 
 	smtp_flush(); /* Flush in every case */
 
+#if 0
+	/* Set the input stream into non-blocking mode */
+	if (pipelining && startrp && smtpfp[0] != NULL)
+	  fd_nonblockingmode(fileno(smtpfp[0]));
+#endif
+
 	time(&env_start); /* Mark the timestamp */
 
 	/* MAIL FROM:<...> -- pipelineable.. */
 	r = smtpwrite(smtpfp, verboselog, 1, SMTPbuf,
 		      pipelining ? startrp : NULL);
 	if (r != EX_OK) {
+	  /* Uh ??  Many new sendmail's have a pathological error mode:
+	        MAIL FROM...
+		451 cannot preopen /etc/aliases.db
+		  (wait a bit)
+		250 ... Sender ok.
+	     We try to accomodate that behaviour, and resync,
+	     although treat it as temporary error -- 4xx series.  */
+	  if (smtp_isopen) {
+	    sleep(10); /* After a sleep of 10 seconds, if we find that
+			  we have some new input, do close the connection */
+	    if (has_readable(fileno(smtpfp[1]))) {
+	      /* Drain the input, and then close the channel */
+	      smtpwrite(smtpfp, verboselog, 1, NULL, NULL);
+	      fclose(smtpfp[0]); smtpfp[0] = NULL;
+	      fclose(smtpfp[1]); smtpfp[1] = NULL;
+	      smtp_isopen = 0;
+	      if (logfp)
+		fprintf(logfp, "%s#\t(closed SMTP channel - MAIL FROM:<> got two responces!)\n", logtag());
+	    }
+	  }
 	  if (pipelining && smtp_isopen /* might have timed out */)
 	    r = smtp_sync(smtpfp, verboselog, r);
 	  for (rp = startrp; rp && rp != endrp; rp = rp->next) {
@@ -1062,7 +1096,7 @@ deliver(dp, smtpfp, verboselog, startrp, endrp)
 	    notaryreport(rp->addr->user,
 			 notaryacct(r,"relayed"), /* XXX: NOTARY DRPT ?? */
 			 NULL,NULL);
-	    diagnostic(rp, r, 0, "%s", remotemsg);
+	    diagnostic(rp, r, 0, "%s", "" /* remotemsg */);
 	  }
 
 	/* More recipients to send ? */
@@ -1518,8 +1552,8 @@ char *buf;
 	    sscanf(buf+5,"%ld",&ehlo_sizeval);
 	} else if (strncmp(buf,"X-RCPTLIMIT",11)==0) {
 	  int nn = atoi(buf+12);
-	  if (nn < 100 || nn > 100000)
-	    nn = 100;
+	  if (nn < 10 || nn > 100000)
+	    nn = 10;
 	  rcpt_limit = nn;
 	}
 	return 0;
@@ -1911,7 +1945,11 @@ fdopen_fail:
 	      if (fd2 < 0 || (fparr[1] = fdopen(fd2, "w")) == NULL)
 		goto fdopen_fail;
 
+
+	      setvbuf(fparr[1], NULL, _IOFBF, smtp_bufsize);
 	      smtp_isopen = 1;
+	      smtp_outcount = 0;
+	      block_written = 0;
 
 	      if (esmtp_on_banner > 0)
 		esmtp_on_banner = 0;
@@ -2011,6 +2049,11 @@ vcsetup(ipaddrcp, ipaddrlen, fdp, hostname, verboselog)
 	  sad.sin_port = htons(servport);
 /* setreuid(0,first_uid);
    if(verboselog) fprintf(verboselog,"setreuid: first_uid=%d, ruid=%d, euid=%d\n",first_uid,getuid(),geteuid()); */
+
+	if (verboselog)
+	  fprintf(verboselog, "Connecting to %s [%s] port %d\n",
+		  hostname, ipaddress, ntohs(sad.sin_port));
+
 	if (connect(s, (struct sockaddr *)&sad, sizeof sad) >= 0) {
 	  int on = 1;
 /* setreuid(0,0); */
@@ -2127,6 +2170,7 @@ int within_ehlo = 0;
  */
 
 static int pipebufsize;
+static int pipebufspace;
 static char *pipebuf = NULL;
 static int pipeindex = 0;
 static int pipespace = 0;
@@ -2137,8 +2181,10 @@ void
 smtp_flush()
 {
 	pipebufsize = 0;
-	if (pipebuf == NULL)
-	  pipebuf = emalloc(1);
+	if (pipebuf == NULL) {
+	  pipebufspace = 240;
+	  pipebuf = emalloc(pipebufspace);
+	}
 	for (;pipeindex > 0; --pipeindex) {
 	  free(pipecmds[pipeindex-1]);
 	}
@@ -2216,12 +2262,36 @@ int fd, timeout;
 	  return 0;
 	return 1;    /* interrupt, or some such.. */
 }
+
+int has_readable(fd)
+int fd;
+{
+	struct timeval tv;
+	int rc;
+	fd_set rdmask;
+
+	tv.tv_sec = 0;
+	tv.tv_usec = 0;
+	_FD_ZERO(rdmask);
+	_FD_SET(fd,rdmask);
+
+	rc = select(fd+1,&rdmask,NULL,NULL,&tv);
+	if (rc == 1) /* There is something to read! */
+	  return 1;
+	return 0;    /* interrupt or timeout, or some such.. */
+}
 #else /* not HAVE_SELECT */
 int select_sleep(fd, timeout)
 int fd, timeout;
 {
 	return -1;
 }
+
+int has_readable(fd)
+int fd;
+{
+	return 1;
+}
 #endif
 
 static int code_to_status(code,statusp)
@@ -2336,7 +2406,13 @@ int smtp_sync(smtpfp,verboselog,r)
 	char *p;
 
 	alarm(timeout);
-	fflush(smtpfp[1]);
+	fflush(smtpfp[1]);			/* Flush output */
+	smtp_outcount = 0;
+	block_written = 0;
+#if 0
+	fd_blockingmode(fileno(smtpfp[0]));	/* Input in blocking
+						   read mode */
+#endif
 	alarm(0);
 
 	s = pipebuf;
@@ -2380,7 +2456,7 @@ int smtp_sync(smtpfp,verboselog,r)
 		fprintf(logfp,"%s#\tRemote gave error %d (%s) while %d responces missing\n",
 			logtag(), err, strerror(err), pipeindex - 1 - idx);
 	      if (verboselog)
-		fprintf(logfp,"Remote gave error %d (%s) while %d responces missing\n",
+		fprintf(verboselog,"Remote gave error %d (%s) while %d responces missing\n",
 			err, strerror(err), pipeindex - 1 - idx);
 	      break;
 	    } else if (len == 0) {
@@ -2389,13 +2465,17 @@ int smtp_sync(smtpfp,verboselog,r)
 		fprintf(logfp,"%s#\tRemote hung up on us while %d responces missing\n",
 			logtag(), pipeindex - idx);
 	      if (verboselog)
-		fprintf(logfp,"Remote hung up on us while %d responces missing\n",
+		fprintf(verboselog,"Remote hung up on us while %d responces missing\n",
 			pipeindex - idx);
 	      break;
 	    } else {
 	      /* more data for processing.. munch munch.. */
 	      eof = pipebuf;
-	      pipebuf = (char*)realloc(pipebuf,pipebufsize+len+1);
+	      if ((pipebufsize+len+1) > pipebufspace) {
+		while ((pipebufsize+len+2) > pipebufspace)
+		  pipebufspace <<= 1; /* Double the size */
+		pipebuf = (char*)realloc(pipebuf,pipebufspace);
+	      }
 	      if (pipebuf != eof) {
 		/* Block changed.. Reset those pointers */
 		long offsetchange = pipebuf - eof;
@@ -2504,11 +2584,11 @@ smtpwrite(smtpfp, verboselog, saverpt, strbuf, syncrp)
 	char buf[8192]; /* XX: Eh... Static buffer.. */
 	char ch;
 
-	alarm(timeout);	/* This much total to write and get answer */
-	gotalarm = 0;
 	/* we don't need to turn alarms off again (yay!) */
 	if (!syncrp) {
-	  /* fflush(smtpfp[0]); */
+	  /*fflush(smtpfp[0]);*/
+	  alarm(timeout);	/* This much total to write and get answer */
+	  gotalarm = 0;
 	  fflush(smtpfp[1]);
 	}
 	infd = FILENO(smtpfp[0]);
@@ -2533,23 +2613,6 @@ smtpwrite(smtpfp, verboselog, saverpt, strbuf, syncrp)
 	  pipercpts[pipeindex] = syncrp;
 	  ++pipeindex;
 
-	  /* Read and buffer all so far accumulated responces.. */
-	  fd_nonblockingmode(infd);
-	  for (;;) {
-	    r = read(infd, buf, sizeof buf);
-	    if (r == 0) break; /* EOF ?! */
-	    if (r < 0)
-	      break;
-	    if (pipebuf == NULL) {
-	      pipebuf = (char*)emalloc(r+1);
-	      pipebufsize = 0;
-	    } else
-	      pipebuf = (char*)erealloc(pipebuf,pipebufsize+r+2);
-	    memcpy(pipebuf+pipebufsize,buf,r);
-	    pipebufsize += r;
-	  }
-	  fd_blockingmode(infd);
-	  /* Continue the processing... */
 	} /* ... end of if(syncrp) */
 
 	if (strbuf != NULL) {
@@ -2563,9 +2626,56 @@ smtpwrite(smtpfp, verboselog, saverpt, strbuf, syncrp)
 	    fwrite(buf, len, 1, verboselog);
 
 	  if (syncrp) {
+	    /* We are asynchronous! */
+	    smtp_outcount += len; /* Where will we grow to ? */
+
+	    if (smtp_outcount > smtp_bufsize) {
+	      /* We will fill the buffer, and flush */
+	      alarm(timeout);	/* This much total to write and get answer */
+	      gotalarm = 0;
+	    }
+
+	    if (block_written && has_readable(infd)) {
+	      /* Read and buffer all so far accumulated responces.. */
+	      fd_nonblockingmode(infd);
+	      for (;;) {
+		r = read(infd, buf, sizeof buf);
+		if (r == 0) break; /* EOF ?! */
+		if (r < 0)
+		  break;
+		if (pipebuf == NULL) {
+		  pipebufspace = 240;
+		  while (pipebufspace < r+1)
+		    pipebufspace <<= 1;
+		  pipebuf = (char*)emalloc(pipebufspace);
+		  pipebufsize = 0;
+		} else {
+		  while (pipebufspace < (pipebufsize+r+2))
+		    pipebufspace <<= 1;
+		  pipebuf = (char*)erealloc(pipebuf,pipebufspace);
+		}
+		memcpy(pipebuf+pipebufsize,buf,r);
+		pipebufsize += r;
+		block_written = 0; /* We drain the accumulated input here,
+				      and can thus mark this draining
+				      unneeded for a while. */
+	      }
+	      fd_blockingmode(infd);
+	      /* Continue the processing... */
+	    }
+
 	    r = fwrite(buf,1,len,smtpfp[1]);
 	    err = (r != len);
+
+	    if (smtp_outcount > smtp_bufsize) {
+	      smtp_outcount -= smtp_bufsize;
+	      block_written = 1;
+	    }
+
 	  } else {
+	    /* We act synchronously */
+	    alarm(timeout);	/* This much total to write and get answer */
+	    gotalarm = 0;
 	    r = write(outfd, buf, len);
 	    err = (r != len);
 	  }
@@ -2627,24 +2737,33 @@ smtpwrite(smtpfp, verboselog, saverpt, strbuf, syncrp)
 	cp = buf;
 
 	if (syncrp) {
-	  /* Read and buffer all so far accumulated responces.. */
-	  fd_nonblockingmode(infd);
-	  for (;;) {
-	    r = read(infd, buf, sizeof buf);
-	    if (r == 0) break; /* EOF ?! */
-	    if (r < 0)
-	      break;
-	    if (pipebuf == NULL) {
-	      pipebuf = (char*)emalloc(r+1);
-	      pipebufsize = 0;
-	    } else
-	      pipebuf = (char*)erealloc(pipebuf,pipebufsize+r+1);
-	    memcpy(pipebuf+pipebufsize,buf,r);
-	    pipebufsize += r;
+	  if (block_written && has_readable(infd)) {
+	    /* Read and buffer all so far accumulated responces.. */
+	    fd_nonblockingmode(infd);
+	    for (;;) {
+	      r = read(infd, buf, sizeof buf);
+	      if (r == 0) break; /* EOF ?! */
+	      if (r < 0)
+		break;
+	      if (pipebuf == NULL) {
+		pipebufspace = 240;
+		while (pipebufspace < r+1)
+		  pipebufspace <<= 1;
+		pipebuf = (char*)emalloc(pipebufspace);
+		pipebufsize = 0;
+	      } else {
+		while (pipebufspace < (pipebufsize+r+2))
+		  pipebufspace <<= 1;
+		pipebuf = (char*)erealloc(pipebuf,pipebufspace);
+	      }
+	      memcpy(pipebuf+pipebufsize,buf,r);
+	      pipebufsize += r;
+	    }
+	    block_written = 0;
+	    fd_blockingmode(infd);
+	    /* .. return to the caller, we have no status except
+	       write timeout.. */
 	  }
-	  fd_blockingmode(infd);
-	  /* .. return to the caller, we have no status except
-	        write timeout.. */
 	  return EX_OK;
 	}
 
