@@ -108,8 +108,9 @@ GDBM_FILE dbmfile;
 	char *s, *t;
 	datum dat;
 	datum key;
+	int rc;
 
-	while (!feof(infile) && !ferror(infile) && !dbm_error(dbmfile)) {
+	while (!feof(infile) && !ferror(infile)) {
 	  *linebuf = 0;
 	  fgets(linebuf,sizeof(linebuf)-1,infile);
 	  if (*linebuf == 0) break; /* Last! */
@@ -140,12 +141,14 @@ GDBM_FILE dbmfile;
 	  dat.dsize = strlen(s)+1;
 
 #ifdef HAVE_NDBM
-	  if (dbm_store(dbmfile, key, dat, DBM_INSERT) != 0)
+	  rc = dbm_store(dbmfile, key, dat, DBM_INSERT);
 #else
-	  if (gdbm_store(dbmfile, key, dat, GDBM_INSERT) != 0)
+	  rc = gdbm_store(dbmfile, key, dat, GDBM_INSERT);
 #endif
-	  {
+	  if (rc > 0) {
 	    fprintf(stderr,"Key: `%s' is duplicate!\n",t);
+	  } else if (rc < 0) {
+	    break; /* Some error -- any, just quit */
 	  }
 	}
 }
